"""
SQLStore to support Kypher queries over KGTK graphs.
"""

import re
from   functools import lru_cache

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *


### Indexing support

# The functions and classes below support the following:
# - extensible representation of arbitrary index objects (such as column, multi-column, text indexes, etc.)
# - support for parsing concise index specs that can be supplied on the command line, for example,
#   '... --idx text:node1,node2/text ...' to specify a full-text search index on a graph column
# - support for storing and retrieving index objects to database info tables
# - support for comparing indexes for equivalence and subsumption
# - support for generating SQL definition/deletion statements specific to a particular index type
# - mapping macro index modes onto their respective index sets or actions
# - TO DO: detect modes such as 'mode:attgraph' automatically from computing some quick statistics

INDEX_MODE_NONE        = 'mode:none'
INDEX_MODE_AUTO        = 'mode:auto'
INDEX_MODE_AUTO_TEXT   = 'mode:autotext'
INDEX_MODE_CLEAR       = 'mode:clear'
INDEX_MODE_CLEAR_TEXT  = 'mode:cleartext'
INDEX_MODE_EXPERT      = 'mode:expert'

# graph modes:
INDEX_MODE_GRAPH       = 'mode:graph'
INDEX_MODE_MONO_GRAPH  = 'mode:monograph'
INDEX_MODE_VALUE_GRAPH = 'mode:valuegraph'
INDEX_MODE_TEXT_GRAPH  = 'mode:textgraph'

# legacy modes:
INDEX_MODE_PAIR        = 'mode:node1+label'
INDEX_MODE_TRIPLE      = 'mode:triple'
INDEX_MODE_QUAD        = 'mode:quad'

INDEX_MODES = {
    # macro modes:
    INDEX_MODE_NONE:        INDEX_MODE_NONE,
    INDEX_MODE_AUTO:        INDEX_MODE_AUTO,
    INDEX_MODE_AUTO_TEXT:   INDEX_MODE_AUTO_TEXT,
    INDEX_MODE_CLEAR:       INDEX_MODE_CLEAR,
    INDEX_MODE_CLEAR_TEXT:  INDEX_MODE_CLEAR_TEXT,
    INDEX_MODE_EXPERT:      INDEX_MODE_EXPERT,

    # graph modes:
    INDEX_MODE_GRAPH:       ['index:node1,label,node2', 'index:label', 'index:node2,label,node1'],
    INDEX_MODE_MONO_GRAPH:  ['index:node1,label,node2', 'index:node2,label,node1'],
    INDEX_MODE_VALUE_GRAPH: ['index:node1'],
    INDEX_MODE_TEXT_GRAPH:  ['index:node1', 'text:node2//tokenize=trigram'],
    
    # legacy modes:
    INDEX_MODE_PAIR:        ['index:node1', 'index:label'],
    INDEX_MODE_TRIPLE:      ['index:node1', 'index:label', 'index:node2'],
    INDEX_MODE_QUAD:        ['index:node1', 'index:label', 'index:node2', 'index:id'],
}

def get_normalized_index_mode(index_spec):
    """Normalize 'index_spec' to one of the legal macro modes such as 'mode:auto', etc.,
    or a list of individual index specs corresponding to the mode.  If 'index_spec' is a
    custom spec such as 'node1,node2', for example, it will also be converted to a list.
    """
    norm_spec = None
    spec_type = get_index_spec_type(index_spec)
    if spec_type and spec_type.lower() == 'mode':
        # we have an explicit mode, look it up and ensure it is valid:
        parse = tokenize_index_spec(index_spec)
        if len(parse) == 2 and parse[1][1] == 'text':
            norm_spec = INDEX_MODES.get('mode:' + parse[1][0].lower())
        if norm_spec is None:
            raise KGTKException(f'unsupported index mode: {index_spec}')
    else:
        # we might have a bare mode such as 'auto' or 'none', try to look it up as a mode
        # (to use a bare mode as a column name, explicitly use the appropriate index type):
        norm_spec = INDEX_MODES.get('mode:' + index_spec.strip().lower(), [index_spec])
        # enforce that clear-modes are fully qualified for some extra protection:
        if norm_spec in (INDEX_MODE_CLEAR, INDEX_MODE_CLEAR_TEXT):
            raise KGTKException(f"index mode '{index_spec}' needs to be explicitly qualified")
    return norm_spec

# we use /<option> as the option syntax instead of the --<option> syntax used on the command line
# for more concise representation, and to visually separate these specs from other command options:
INDEX_TOKENIZER_REGEX = re.compile(
    '|'.join([r'(?P<optsepsep>//)\s*',            # '//' (needs to come before single '/')
              r'(?P<optsep>/)\s*',                # '/'
              r'(?P<typesep>:)\s*',               # ':'
              r'(?P<valuesep>=)\s*',              # '='
              r'(?P<sep>[,()])\s*',               # (',', '(', ')')
              r'(?P<text>[^,()/:=`"\s]+)',        # non-special-char text tokens
              r'`(?P<quote_1>([^`]*``)*[^`]*)`',  # `-quoted tokens
              r'"(?P<quote_2>([^"]*"")*[^"]*)"',  # "-quoted tokens
              r'(?P<whitespace>\s+)',             # whitespace separates but is ignored
    ]))

INDEX_SPEC_TYPE_SEPARATOR = ':'

@lru_cache(maxsize=None)
def tokenize_expression(expression, regex=INDEX_TOKENIZER_REGEX):
    """Tokenize expression into a list of '(token, type)' tuples where type is one of 'text' or 'sep'.
    Tokens are split at separators and whitespace unless prevented by quoting (all defined by 'regex').
    Quoting is performed just like identifier quoting in SQL or Cypher using either a backtick or
    double quote where an explicit quote can be inserted by doubling it.
    """
    tokens = []
    total_match = 0
    for m in regex.finditer(expression):
        ms, me = m.span()
        token = m.group(m.lastgroup)
        toktype = m.lastgroup.split('_')[0]
        total_match += (me - ms)
        if toktype == 'quote':
            quote = expression[ms]
            # unescape quotes:
            token = token.replace(quote+quote, quote)
            toktype = 'text'
        if toktype != 'whitespace':
            tokens.append((token, toktype))
    # make sure we didn't skip any garbage:
    if total_match < len(expression):
        raise KGTKException('illegal expression syntax')
    return tokens

def tokenize_index_spec(index_spec, regex=INDEX_TOKENIZER_REGEX):
    """Tokenizes 'index_spec' (unless it is already tokenized) and returns all
    text tokens classified as one of ('text', 'type', 'option', 'global-option').
    All separator tokens are interpreted and then filtered out.
    """
    if not isinstance(index_spec, list):
        index_spec = tokenize_expression(index_spec, regex=regex)
    index_spec = [list(x) for x in index_spec]
    last = len(index_spec) - 1
    tokens = []
    for i, (token, toktype) in enumerate(index_spec):
        if toktype == 'typesep':
            if i > 0 and index_spec[i-1][1] == 'text':
                index_spec[i-1][1] = 'type'
            else:
                raise KGTKException('illegal index spec syntax')
        elif toktype in ('optsep', 'optsepsep'):
            if i < last and index_spec[i+1][1] == 'text':
                index_spec[i+1][1] = 'option' if toktype == 'optsep' else 'global-option'
            else:
                raise KGTKException('illegal index spec syntax')
        elif toktype == 'valuesep':
            value = '' # an option followed by non-text means the empty value
            if i < last and index_spec[i+1][1] == 'text':
                value = index_spec[i+1][0]
                index_spec[i+1][1] = 'value'
            if i > 0 and index_spec[i-1][1].endswith('option'):
                # if we have a value, we represent it with a tuple:
                index_spec[i-1][0] = (index_spec[i-1][0], value)
            else:
                raise KGTKException('illegal index spec syntax')
    for token, toktype in index_spec:
        if toktype in ('text', 'type', 'option', 'global-option'):
            tokens.append((token, toktype))
    return tokens

def get_index_spec_type(index_spec):
    """Return 'index_spec's type if it starts with one, otherwise return None.
    This will also return None for some syntatically incorrect specs, but these
    errors should be caught during full parsing of the spec.
    """
    seppos = index_spec.find(INDEX_SPEC_TYPE_SEPARATOR)
    if seppos >= 0:
        tokens = tokenize_expression(index_spec[0:seppos+1])
        if len(tokens) == 2 and tokens[0][1] == 'text' and tokens[1][1] == 'typesep':
            return tokens[0][0]
    return None

def parse_index_spec(index_spec, regex=INDEX_TOKENIZER_REGEX):
    """Parse 'index_spec' (a string or tokenized list) into an initial sdict representation.
    Local and global option values are parsed and appropriately assigned.  Index-specific
    'parse_spec' methods can do any further normalizations if necessary.
    """
    tokens = tokenize_index_spec(index_spec, regex=regex)
    parse = sdict['type': None, 'columns': sdict(), 'options': {}]
    column_options = None
    for (token, toktype) in tokens:
        if toktype == 'text':
            column_options = {}
            parse.columns[token] = column_options
        elif toktype in ('option', 'global-option'):
            opt, value = (token, True) if isinstance(token, str) else token
            try:
                import ast
                if value in ('true', 'TRUE', 'false', 'FALSE'):
                    value = value.capitalize()
                value = ast.literal_eval(value) # dwim booleans and numbers
            except:
                pass                            # everything else is considered a string
            if toktype == 'global-option':
                parse.options[opt] = value
            elif column_options is not None:
                column_options[opt] = value
            else:
                raise KGTKException('illegal index spec syntax')
        elif toktype == 'type':
            if parse.type is None and len(parse.columns) == 0 and len(parse.options) == 0:
                parse.type = token
            else:
                raise KGTKException('illegal index spec syntax')
        else:
            raise KGTKException('index spec parsing error')
    return parse


class TableIndex(object):
    """Represents objects to describe and manipulate database table indexes (aka indices).
    """

    def __init__(self, table, index_spec):
        """Create an index object for 'table' based on 'index_spec' which can be
        represented as an sdict object, string version of an sdict object, or valid
        and parsable index_spec short form (.e.g., 'index: node1, node2').
        """
        self.table = table
        self.index = index_spec
        self.index = self.get_index()

    def __repr__(self):
        """Create an eval-able repr that will recreate 'self' identically.
        """
        return f"{type(self).__name__}({repr(self.table)}, {self.index})"

    def __eq__(self, other):
        return (type(self) == type(other)
                and self.index == other.index
                and self.get_table_name() == other.get_table_name())

    @classmethod
    def encode(self, index_tree):
        """Return a string encoding of 'index_tree' that can be stored to the DB.
        """
        return repr(index_tree)

    @classmethod
    def decode(self, index_expr):
        """Convert 'index_expr' (a string encoding of an index tree created by 'encode')
        back into the corresponding index object(s).
        """
        return eval(index_expr)

    def get_index(self):
        """Return the parsed index for 'self'.
        """
        index = self.index
        if isinstance(index, sdict):
            pass
        elif isinstance(index, str):
            if index.startswith('sdict['):
                index = eval(index)
            else:
                index = self.parse_spec(index)
        else:
            raise KGTKException(f'illegal index spec: {index}')
        self.index = index
        if type(self).__name__ != self.INDEX_TYPES.get(index.type):
            # minor hackery to instantiate to the right class depending on the index spec:
            if index.type not in self.INDEX_TYPES:
                raise KGTKException(f'unsupported index spec type: {index.type}')
            klass = eval(self.INDEX_TYPES[index.type])
            # change-class (pretend we're in Lisp):
            self.__class__ = klass
        return index

    INDEX_TYPES = {'index': 'StandardIndex', 'text': 'TextIndex', 'sql': 'SqlIndex', 'vector': 'VectorIndex'}
    DEFAULT_INDEX_TYPE = 'index'

    def get_index_type_name(self):
        return next(k for k,v in self.INDEX_TYPES.items() if v == self.__class__.__name__)

    @classmethod
    def get_index_type_class(self, index_type):
        class_name = self.INDEX_TYPES.get(index_type)
        if class_name is None:
            raise KGTKException(f'unsupported index spec type: {index_type}')
        else:
            return eval(class_name)

    def parse_spec(self, index_spec):
        """Parse a short-form string 'index_spec' and return the result as an sdict.
        This simply dispatches to the appropriate index subclasses.
        """
        spec_type = get_index_spec_type(index_spec) or self.DEFAULT_INDEX_TYPE
        klass = self.get_index_type_class(spec_type)
        return klass(self.table, index_spec).index

    def get_table_name(self):
        if hasattr(self.table, '_name_'):
            return self.table._name_
        elif isinstance(self.table, str):
            return self.table
        else:
            raise KGTKException('illegal table type')

    def get_name(self):
        """Return the SQL name to be used for this index
        """
        raise KGTKException('not implemented')

    def get_create_script(self):
        """Return a list of SQL statements required to create this index.
        """
        raise KGTKException('not implemented')

    def get_drop_script(self):
        """Return a list of SQL statements required to delete this index.
        """
        raise KGTKException('not implemented')

    def create_index(self, store, explain=False):
        """Perform all actions needed to create this index on 'store'.
        If 'explain' only describe what needs to be done without actually doing it.
        """
        loglevel = 0 if explain else 1
        for index_stmt in self.get_create_script():
            store.log(loglevel, index_stmt)
            if not explain:
                store.execute(index_stmt)
                
    def drop_index(self, store, explain=False):
        """Perform all actions needed to delete this index on 'store'.
        If 'explain' only describe what needs to be done without actually doing it.
        """
        loglevel = 0 if explain else 1
        for index_stmt in self.get_drop_script():
            store.log(loglevel, index_stmt)
            if not explain:
                store.execute(index_stmt)

    def has_primary_column(self, column):
        """Return True if this index has 'column' as its first indexed column.
        """
        for key in self.index.columns.keys():
            return key == column

    def subsumes_columns(self, columns):
        """Return True if 'columns' are a prefix of this index's columns,
        that is, it might handle a superset of lookup requests.  Note that
        this does not consider the type of index or any options such as 'unique',
        so actual subsumption is determined only by the respective 'subsumes'.
        """
        index_columns = self.index.columns.keys()
        columns = [columns] if isinstance(columns, str) else columns
        for idx_column, column in zip(index_columns, columns):
            if idx_column != column:
                return False
        return len(columns) <= len(index_columns)

    def subsumes(self, index):
        """Return True if 'self' subsumes or is more general than 'index',
        that is it can handle a superset of lookup requests.
        This does not (yet) consider any options such as 'unique'.
        """
        return self.table == index.table and self.subsumes_columns(index.index.columns.keys())

    def redefines(self, index):
        """Return True if 'self' is different from 'index' and redefines it.
        """
        return False

    
class StandardIndex(TableIndex):
    """Standard column indexes created via 'CREATE INDEX...'.
    """

    def parse_spec(self, index_spec):
        """Parse a standard table 'index_spec' such as, for example:
        'index: node1, label, node2 //unique' or 'node1, label, node2' 
        ('index' is the default index spec type if not supplied).
        """
        parse = parse_index_spec(index_spec)
        type_name = self.get_index_type_name()
        if parse.type is None:
            parse.type = type_name
        if parse.type != type_name:
            raise KGTKException(f'mismatched index spec type: {parse.type}')
        return parse

    def get_name(self):
        """Return the global SQL name to be used for this index.
        """
        table_name = self.get_table_name()
        column_names = '_'.join(self.index.columns.keys())
        index_name = '%s_%s_idx' % (table_name, column_names)
        return index_name

    def get_create_script(self):
        """Return a list of SQL statements required to create this index.
        """
        table_name = self.get_table_name()
        index_name = self.get_name()
        options = self.index.options
        columns = self.index.columns
        column_names = list(columns.keys())
        unique = 'UNIQUE ' if options.get('unique', False) or columns[column_names[0]].get('unique', False) else ''
        column_names = ', '.join([sql_quote_ident(col) for col in column_names])
        statements = [
            f'CREATE {unique}INDEX {sql_quote_ident(index_name)} ON {sql_quote_ident(table_name)} ({column_names})',
            # do this unconditionally for now, given that it only takes about 10% of index creation time:
            f'ANALYZE {sql_quote_ident(index_name)}',
        ]
        return statements

    def get_drop_script(self):
        """Return a list of SQL statements required to delete this index.
        """
        statements = [
            f'DROP INDEX {sql_quote_ident(self.get_name())}'
        ]
        return statements

    def subsumes(self, index):
        """Return True if 'self' subsumes or is more general than 'index',
        that is it can handle a superset of lookup requests.
        This does not (yet) consider any options such as 'unique'.
        """
        return (self.table == index.table
                and isinstance(index, (StandardIndex, SqlIndex))
                and self.subsumes_columns(index.index.columns.keys()))


# TextIndex NOTES:
# - all columns will be indexed unless excluded with 'unindexed'
# - tables are contentless, since we need to match to the source table via rowid anyway
# - trigram seems to be the most powerful tokenizer, so we use that as the default, however,
#   it uses extra space, and it requires SQLite 3.34.0 which requires Python 3.9 or later
# - we support optional names on indexes, which allows us to easily redefine them and to
#   have multiple indexes on the same source
# - indexing scores are between -20 and 0, if we rerank with pagerank, that needs to be
#   considered, for example, additive weighting with log(pagerank) seems like an option
# - matching on node IDs works too and doesn't require special tokenizer options
# - we should have a //strip or //preproc option to specify a custom preprocessing function

# Index/tokenizer performance tradeoffs:
# - case-insensitive trigram (default): fast textmatch, fast textlike, fast textglob, more space
# - case-sensitive trigram: fast textmatch, fast textglob, no textlike, more space than case-insensitive
# - ascii, unicode61: fast textmatch on whole words, also prefixes if //prefix is specified,
#   no textlike, no textglob, less space
    
class TextIndex(TableIndex):
    """Specialized indexes to support full-text search via SQLite's FT5.
    """

    COLUMN_OPTIONS    = ('unindexed')
    TABLE_OPTIONS     = ('tokenize', 'prefix', 'content', 'columnsize', 'detail', 'name')
    
    DEFAULT_TOKENIZER = 'trigram'
    # not all of these apply to all tokenizers, but we don't model that for now:
    TOKENIZE_OPTIONS  = ('categories', 'tokenchars', 'separators', 'remove_diacritics', 'case_sensitive')
    
    def parse_spec(self, index_spec):
        """Parse a full-text 'index_spec' such as, for example:
        'text:node1/unindexed,node2//name=labidx//prefix=2//tokenize=trigram'
        The 'unindexed' option should be rare and is just shown for illustration.
        """
        parse = parse_index_spec(index_spec)
        if parse.type != self.get_index_type_name():
            raise KGTKException(f'mismatched index spec type: {parse.type}')
        for key in parse.options.keys():
            if not (key in self.TABLE_OPTIONS or key in self.TOKENIZE_OPTIONS):
                raise KGTKException(f'unhandled text index option: {key}')
        if 'tokenize' not in parse.options:
            for subopt in self.TOKENIZE_OPTIONS:
                if parse.options.get(subopt) is not None:
                    raise KGTKException(f'missing tokenize option for {subopt}')
        # use content-less indexes linked to graph by default (override with //content):
        content = parse.options.get('content')
        if not content:    # None, False, ''
            parse.options['content'] = self.get_table_name()
        elif content is True:
            del parse.options['content']
        return parse

    def get_name(self):
        """Return the global SQL name to be used for this index.
        """
        table_name = self.get_table_name()
        index_name = self.index.options.get('name')
        if index_name is None:
            import shortuuid
            # generate a name based on the index itself instead of external state
            # (minor gamble on uniqueness with shortened key):
            index_name = shortuuid.uuid(repr(self)).lower()[0:10] + '_'
        return f'{table_name}_txtidx_{index_name}'

    def get_create_script(self):
        """Return a list of SQL statements required to create this index.
        """
        table_name = self.get_table_name()
        index_name = self.get_name()
        columns = self.index.columns
        column_names = ', '.join([sql_quote_ident(col) for col in columns.keys()])
        column_names_with_options = ', '.join(
            [sql_quote_ident(col) + (' UNINDEXED' if columns[col].get('unindexed', False) else '')
             for col in columns.keys()])
        
        options = self.index.options
        index_options = []
        if 'tokenize' in options:
            tokopt = str(options.get('tokenize'))
            for subopt in self.TOKENIZE_OPTIONS:
                value = options.get(subopt)
                if value is not None:
                    value = str(int(value)) if isinstance(value, bool) else str(value)
                    tokopt += f""" {subopt} {sql_quote_ident(value, "'")}"""
            tokopt = f"""tokenize={sql_quote_ident(tokopt)}"""
            index_options.append(tokopt)
        else:
            tokopt = f"""tokenize={sql_quote_ident(self.DEFAULT_TOKENIZER)}"""
            index_options.append(tokopt)
        if 'prefix' in options:
            index_options.append(f"""prefix={sql_quote_ident(str(options.get('prefix')))}""")
        if 'content' in options:
            index_options.append(f"""content={sql_quote_ident(str(options.get('content')))}""")
        if 'columnsize' in options:
            index_options.append(f"""columnsize={sql_quote_ident(str(options.get('columnsize')))}""")
        if 'detail' in options:
            index_options.append(f"""detail={options.get('detail')}""")
        if index_options:
            column_names_with_options += (', ' + ', '.join(index_options))
        
        statements = [
            f'CREATE VIRTUAL TABLE {sql_quote_ident(index_name)} USING FTS5 ({column_names_with_options})',
            f'INSERT INTO {sql_quote_ident(index_name)} ({column_names}) SELECT {column_names} FROM {table_name}',
        ]
        return statements

    def get_drop_script(self):
        """Return a list of SQL statements required to delete this index.
        """
        statements = [
            f'DROP TABLE {sql_quote_ident(self.get_name())}'
        ]
        return statements

    def subsumes(self, index):
        """Return True if 'self' subsumes or is more general than 'index',
        that is it can handle a superset of lookup requests.
        """
        # for now we require strict equivalence:
        return self == index

    def redefines(self, index):
        """Return True if 'self' is different from 'index' and redefines it.
        Text indexes redefine based on a defined and equal name to another text index.
        """
        return (isinstance(index, TextIndex)
                and self != index
                and self.index.options.get('name') is not None
                and self.index.options['name'] == index.index.options.get('name'))

    
class SqlIndex(TableIndex):
    """Handle SQL CREATE INDEX statements.
    """

    def parse_spec(self, index_spec):
        """Parse an SQL 'index_spec' such as, for example:
        'sql: CREATE UNIQUE INDEX "graph_1_node1_idx" on graph_1 ("node1")'
        This supports the subset of creation statement this module produces.
        """
        tokens = tokenize_expression(index_spec)
        type_name = self.get_index_type_name()
        if get_index_spec_type(index_spec) != type_name:
            raise KGTKException(f'not an SQL index spec: {index_spec}')
        definition = index_spec[index_spec.find(':')+1:].strip()
        parse = sdict['type': type_name, 'columns': sdict(), 'options': {}, 'definition': definition]
        tokens = tokens[2:]
        tokens = list(reversed(tokens))
        try:
            if tokens.pop()[0].upper() != 'CREATE':
                raise Exception()
            token = tokens.pop()[0].upper()
            if token == 'UNIQUE':
                parse.options['unique'] = True
                token = tokens.pop()[0].upper()
            if token != 'INDEX':
                raise Exception()
            # 'IF NOT EXISTS' would go here:
            parse.options['name'] = tokens.pop()[0]
            if tokens.pop()[0].upper() != 'ON':
                raise Exception()
            parse.options['table'] = tokens.pop()[0]
            if tokens.pop()[0] != '(':
                raise Exception()
            column_options = None
            for token, toktype in reversed(tokens):
                tokens.pop()
                if token == ')':
                    break
                if toktype == 'text':
                    if column_options is None:
                        column_options = {}
                        parse.columns[token] = column_options
                    else:
                        # 'COLLATION', 'ASC', 'DESC' would go here:
                        raise Exception()
                elif token == ',' and toktype == 'sep':
                    column_options = None
                else:
                    raise Exception()
            if len(tokens) > 0:
                raise Exception()
        except:
            raise KGTKException(f'illegal or unhandled SQL index spec: {index_spec}')
        if self.table is not None and self.table != parse.options['table']:
            raise KGTKException(f'table in index object does not match index definition')
        return parse

    def get_table_name(self):
        if self.table is None:
            return self.index.options['table']
        else:
            return super().get_table_name()

    def get_name(self):
        """Return the global SQL name to be used for this index.
        """
        return self.index.options['name']

    def get_create_script(self):
        """Return a list of SQL statements required to create this index.
        """
        return [self.index.definition,
                f'ANALYZE {sql_quote_ident(self.get_name())}',
        ]

    def get_drop_script(self):
        """Return a list of SQL statements required to delete this index.
        """
        statements = [
            f'DROP INDEX {sql_quote_ident(self.get_name())}'
        ]
        return statements

    def subsumes(self, index):
        """Return True if 'self' subsumes or is more general than 'index',
        that is it can handle a superset of lookup requests.
        This does not (yet) consider any options such as 'unique'.
        """
        return (self.table == index.table
                and isinstance(index, (SqlIndex, StandardIndex))
                and self.subsumes_columns(index.index.columns.keys()))


class VectorIndex(TableIndex):
    """Specialized indexes needed for numeric vectors.  Since these are not really
    SQL indexes, this abuses the current SQL table-centric model a little bit,
    however, there is some SQL support needed as well so it is not all wrong.
    """

    FORMAT_AUTO    = 'auto'
    FORMAT_TEXT    = 'text'
    FORMAT_BASE64  = 'base64'
    DTYPE_FLOAT32  = 'float32'
    DTYPE_FLOAT16  = 'float16'
    DTYPE_FLOAT64  = 'float64'
    NORM_L2        = 'l2'
    NN_INDEX_FAISS = 'faiss'
    STORE_INLINE   = 'inline'
    STORE_NUMPY    = 'numpy'
    STORE_HD5      = 'hd5'

    DEFAULT_COLUMN = 'node2'
    # the first value in each option list below is the default:
    COLUMN_OPTIONS = {
        # 'fmt' specifies the format the vectors are in, which can either be a plain text list of
        #  numbers separated by one of ' ,;:|', or a base64 encoding of a numpy <vector>.tobytes();
        # 'auto' will try to guess the format from the first vector in the imported list:
        'fmt':   (FORMAT_AUTO, FORMAT_TEXT, FORMAT_BASE64,),
        # 'dtype' specifies a numpy element data type to use for the imported vectors:
        'dtype': (DTYPE_FLOAT32, DTYPE_FLOAT16, DTYPE_FLOAT64,),
        # if 'norm' is requested, vectors will be normalized before they are stored, but the
        # norm will be stored as well so it can be used to unnormalize a vector if needed
        # (for now we only support the L2 norm):
        'norm':  (None, True, False, NORM_L2,),
        # 'store' controls how imported vectors should be stored (only 'inline' for now):
        'store': (STORE_INLINE, STORE_NUMPY, STORE_HD5,),
        # 'ext' specifies an external existing file to use for a Numpy store:
        'ext': (None, str),
        # 'nn' controls whether a nearest neighbor index should be built, and of what kind:
        'nn':    (False, True, NN_INDEX_FAISS,),
        # 'ram' specifies the maximum amount of RAM to use when training an NN index:
        'ram': (None, str, int),
        # 'nlist' is the number of inverted lists (or cells/centroids) to use for the NN index:
        'nlist': (None, str, int),
        # 'niter' specifies how many iterations to use when training the NN index quantizer:
        'niter': (None, int),
        # 'nprobe' specifies how many quantizer cells to search by default for NN-lookup:
        'nprobe': (None, int),
    }

    def parse_spec(self, index_spec):
        """Parse a vector table 'index_spec' such as, for example:
        'vector:node2' or 'vector:node2/fmt=base64/nn=faiss,node1;txtemb/dtype=float16'.
        """
        parse = parse_index_spec(index_spec)
        type_name = self.get_index_type_name()
        if parse.type is None:
            parse.type = type_name
        if parse.type != type_name:
            raise KGTKException(f'mismatched index spec type: {parse.type}')
        if len(parse.columns) == 0:
            # use the default column if nothing is specified:
            parse['columns'] = sdict[self.DEFAULT_COLUMN: {}]
        for column, options in list(parse.columns.items()):
            for opt, val in list(options.items()):
                if opt.lower() not in self.COLUMN_OPTIONS.keys():
                    raise KGTKException(f'unhandled vector index option: {opt}')
                legal_values = self.COLUMN_OPTIONS[opt.lower()]
                if val not in legal_values and str(val).lower() not in legal_values and type(val) not in legal_values:
                    raise KGTKException(f'unhandled vector option value for {opt}: {val}')
                del options[opt]
                if isinstance(val, str) and type(val) not in legal_values:
                    val = val.lower()
                options[opt.lower()] = val
            for opt in self.COLUMN_OPTIONS.keys():
                if opt not in options:
                    options[opt] = self.COLUMN_OPTIONS[opt][0]
            # map True onto respective default values:
            if options['nn']:
                options['nn'] = self.NN_INDEX_FAISS
            # NOTE: normalization on NumPy external files is specially handled by the vector store
            if options['ext'] and options['store'] != self.STORE_NUMPY:
                raise KGTKException('external files only supported for NumPy vector stores')
            for sopt in ('ram', 'nlist'):
                if options[sopt]:
                    options[sopt] = parse_memory_size(options[sopt])
            # convert column options to sdict so we can use property syntax:
            parse.columns[column] = sdict(options)
        # we need to have exactly one vector index per indexed vector column,
        # and we want separate index objects if there are multiple vector columns:
        if len(parse.columns) != 1:
            raise KGTKException('each vector index must specify exactly 1 column')
        return parse

    def get_name(self):
        """Return the global name to be used for this index (not really used).
        """
        table_name = self.get_table_name()
        index_name = self.index.options.get('name')
        if index_name is None:
            import shortuuid
            # generate a name based on the index itself instead of external state
            # (minor gamble on uniqueness with shortened key):
            index_name = shortuuid.uuid(repr(self)).lower()[0:10] + '_'
        return f'{table_name}_vecidx_{index_name}'

    def get_create_script(self):
        """Return a list of SQL statements required to create this index.
        Since this is not an SQL index, we return the empty list here
        and instead do everything inside 'create_index()'.
        """
        return []

    def get_drop_script(self):
        """Return a list of SQL statements required to delete this index.
        Since this is not an SQL index, we return the empty list here
        and instead do everything inside 'drop_index()'.
        """
        return []

    def create_index(self, store, explain=False):
        """Perform all actions needed to create this index on 'store'.
        If 'explain' only describe what needs to be done without actually doing it.
        """
        loglevel = 0 if explain else 1
        store.log(loglevel, f'CREATE VECTOR INDEX {self.index}...')
        master_store = store.get_vector_store()
        table = self.get_table_name()
        column = list(self.index.columns.keys())[0]
        # this will redefine 'vstore' in case the index spec changed in a redefinition:
        vstore = master_store.get_vector_store(table, column, index_spec=self)
        nnindex = vstore.get_nearest_neighbor_index()
        if nnindex:
            nnindex.create(force=False, explain=explain)
                
    def drop_index(self, store, explain=False):
        """Perform all actions needed to delete this index on 'store'.
        If 'explain' only describe what needs to be done without actually doing it.
        """
        loglevel = 0 if explain else 1
        store.log(loglevel, f'DROP VECTOR INDEX {self.index}...')
        master_store = store.get_vector_store()
        table = self.get_table_name()
        column = list(self.index.columns.keys())[0]
        vstore = master_store.get_vector_store(table, column)
        nnindex = vstore.get_nearest_neighbor_index()
        if nnindex:
            nnindex.drop(explain=explain)
                
    def subsumes(self, index):
        """Return True if 'self' subsumes or is more general than 'index',
        that is it can handle a superset of lookup requests.
        """
        # we require strict equivalence:
        return self == index

    def redefines(self, index):
        """Return True if 'self' is different from 'index' and redefines it.
        Vector indexes need to be about the same vector column for redefinition to occur.
        """
        return (isinstance(index, VectorIndex)
                and self.index.columns.keys() == index.index.columns.keys()
                and self != index)

    def redefines_store(self, index):
        """Return True if 'self' is different from 'index' and redefines its vector storage options.
        """
        if self.redefines(index):
            for column1, options1 in self.index.columns.items():
                options2 = index.index.columns[column1]
                if options1.store != options2.store or options1.dtype != options2.dtype or options1.norm != options2.norm:
                    return True
        return False

    def redefines_quantizer(self, index):
        """Return True if 'self' is different from 'index' but redefines its quantizer index only.
        """
        return self.redefines(index) and not self.redefines_store(index)
    

"""
>>> TableIndex('graph1', 'node1, label, node2')
StandardIndex('graph1', sdict['type': 'index', 'columns': sdict['node1': {}, 'label': {}, 'node2': {}], 'options': {}])
>>> _.get_create_script()
['CREATE INDEX "graph1_node1_label_node2_idx" ON "graph1" ("node1", "label", "node2")',
 'ANALYZE "graph1_node1_label_node2_idx"']

>>> TableIndex('graph2', 'text:node1,node2//tokenize=trigram//case_sensitive//name=myidx')
TextIndex('graph2', sdict['type': 'text', 'columns': sdict['node1': {}, 'node2': {}], 'options': {'tokenize': 'trigram', 'case_sensitive': True, 'name': 'myidx', 'content': 'graph2'}])
>>> _.get_create_script()
['CREATE VIRTUAL TABLE "graph2_txtidx_myidx" USING FTS5 ("node1", "node2", tokenize="trigram case_sensitive \'1\'", content="graph2")', 
 'INSERT INTO "graph2_txtidx_myidx" ("node1", "node2") SELECT "node1", "node2" FROM graph2']

>>> TableIndex('graph_1', 'sql: CREATE UNIQUE INDEX "graph_1_node1_idx" on graph_1 ("node1")')
SqlIndex('graph_1', sdict['type': 'sql', 'columns': sdict['node1': {}], 'options': {'unique': True, 'name': 'graph_1_node1_idx', 'table': 'graph_1'}, 'definition': 'CREATE UNIQUE INDEX "graph_1_node1_idx" on graph_1 ("node1")'])
>>> _.get_create_script()
['CREATE UNIQUE INDEX "graph_1_node1_idx" on graph_1 ("node1")',
 'ANALYZE "graph_1_node1_idx"']

>>> TableIndex('graph2', 'vector:node2/fmt=base64/nn=faiss/store=inline')
VectorIndex('graph2', sdict['type': 'vector', 'columns': sdict['node2': sdict['fmt': 'base64', 'nn': 'faiss', 'store': 'inline', 'dtype': 'float32', 'norm': None, 'ext': None, 'ram': None, 'nlist': None, 'niter': None, 'nprobe': None]], 'options': {}])
# we need a separate spec if we want to index a different column:
>>> TableIndex('graph2', 'vector:node1;txtemb/dtype=float16/norm=False')
VectorIndex('graph2', sdict['type': 'vector', 'columns': sdict['node1;txtemb': sdict['dtype': 'float16', 'norm': False, 'fmt': 'auto', 'store': 'inline', 'ext': None, 'nn': False, 'ram': None, 'nlist': None, 'niter': None, 'nprobe': None]], 'options': {}])
>>> _.get_create_script()
[]
>>> 
"""



"""
SQL function definition API
"""

import sqlite3
import copy

from   kgtk.exceptions import KGTKException


### SqlFunction API

# Provides enhanced functionality compared to original API such as:
# - forward-declarations for lazy loading of functions and support modules
# - @sqlfun decorator for easy definition
# - customizable call-translation API via 'translate_call_to_sql' needed
#   by funcions such as 'cast' or 'likelihood' or the new vector functions
# - customizable code-loading API that allows the same function to be associated
#   with different call-specific codes and closure variables (also primarily
#   needed to properly optimize vector function calls)
#
# We basically support three aspects for each SQL function:
# - definition
# - translation (default or specialized to a function or call context)
# - code generation and registration (default or specialized to a call context)
#
# TO DO:
# - handle multi-threading: the call-specific code generation should support
#   this for multiple threads as well, but the details need to be fleshed out


class SqlFunction(object):

    def __init__(self, name=None, code=None, num_params=None, deterministic=False):
        """Create a function object with 'name', 'code', specify its number of
        parameters with 'num_params' and specify whether it is 'deterministic'
        (those last three parameters mirror 'sqlite3.Cursor.create_function').
        If no 'name' is supplied it will be inferred from 'code', but at least
        one of the two needs to be supplied.  If 'num_params' is left unspecified,
        it will be inferred at code registration time which will require loading
        the 'inspect' module, so it should generally be specified to minimize
        the number of imported modules.
        """
        if name is None and code is not None:
            name = code.__name__
        if name is None:
            raise KGTKException(f'cannot determine name of SQL function')
        self.name = name
        self.code = code
        self.num_params = num_params
        self.deterministic = deterministic
        self.store = None
        self.uniquify = False

    _definitions = {}
    
    def define(self):
        """Add this instance under its current name to the global registry.
        This allows the same function definition to be used as different variants
        under different names (see also 'uniquify_name').
        """
        SqlFunction._definitions[self.get_name()] = self
        return self

    @staticmethod
    def is_defined(name):
        """Return True if a function with 'name' is already defined or forward-declared.
        """
        name = SqlFunction.normalize_name(name)
        return SqlFunction._definitions.get(name) is not None

    @staticmethod
    def declare(module_name, *func_name):
        """Declare module 'module_name' as the one containing definitions for all
        the listed function names 'func_name'.  This allows us to properly translate
        function calls without having to eagerly import them all.
        """
        for func in func_name:
            func = SqlFunction.normalize_name(func)
            if func not in SqlFunction._definitions:
                SqlFunction._definitions[func] = module_name

    @staticmethod
    def import_declared_function(name):
        """Import the forward-declared function 'name' and return it.
        Raise an error if the function is not forward-declared or cannot be found.
        """
        nname = SqlFunction.normalize_name(name)
        module = SqlFunction._definitions.get(nname)
        if not isinstance(module, str):
            raise KGTKException(f'not a forward-declared SQL function: {name}')
        # we have a proper forward-declaration to a defining module, import it:
        exec(f'import {module}')
        fun = SqlFunction._definitions.get(nname)
        if not isinstance(fun, SqlFunction):
            raise KGTKException(f'missing definition for declared SQL function: {name}')
        return fun

    @staticmethod
    def get_function(name, store=None, error=True, **kwargs):
        """Lookup a customizable copy of the function object for 'name'.
        If 'name' is undefined an error will be raised unless 'error' is False.
        The resulting object will be a new copy of the definition object which
        will not be shared by any subsequent calls to this or other functions.
        Assign 'store' as the SQL store associated with the resulting object
        and add any additional 'kwargs' as custom slots to the object.
        """
        fun = SqlFunction._definitions.get(SqlFunction.normalize_name(name))
        if fun is None:
            if error:
                raise KGTKException(f'undefind SQL function: {name}')
            return None
        elif isinstance(fun, str):
            # we have a forward-declaration to a defining module, import it
            # (any errors here are real and not subject to the 'error' flag):
            fun = SqlFunction.import_declared_function(name)
        # create a copy of the definition with some additional values filled in:
        fun = copy.copy(fun)
        fun.store = store
        for key, value in kwargs.items():
            setattr(fun, key, value)
        return fun

    @staticmethod
    def is_aggregate(name):
        """Return True if a function with 'name' is defined or forward-declared
        as an aggregation function.
        """
        # we don't use an instance method for this test, since we need to be able
        # to test this before we actually have an object in hand:
        nname = SqlFunction.normalize_name(name)
        fun = SqlFunction._definitions.get(nname)
        if isinstance(fun, str):
            # we have a forward-declaration to a defining module, import it:
            fun = SqlFunction.import_declared_function(name)
        return isinstance(fun, AggregateFunction)

    @staticmethod
    def is_virtual_graph(name):
        """Return True if a function with 'name' is defined or forward-declared
        as a virtual graph function.
        """
        # we don't use an instance method for this test, since we need to be able
        # to test this before we actually have an object in hand:
        nname = SqlFunction.normalize_name(name)
        fun = SqlFunction._definitions.get(nname)
        if isinstance(fun, str):
            # we have a forward-declaration to a defining module, import it:
            fun = SqlFunction.import_declared_function(name)
        return isinstance(fun, VirtualGraphFunction)
    
    @staticmethod
    def normalize_name(name):
        return name.lower()
    
    def get_name(self):
        """Return the normalized name of this function.
        """
        return self.normalize_name(self.name)

    def get_code(self):
        """Return the code object of this function.  This can return custom adaptations
        relative to a particular function or call context.
        """
        if self.code is None:
            raise KGTKException(f'missing code object for SQL function: {self.name}')
        return self.code

    def get_num_params(self):
        """Return the number of parameters expected by this function (-1 for variable).
        If not previously specified, infer it from the code object via 'inspect'.
        """
        if self.num_params is None:
            import inspect
            argspec = inspect.getfullargspec(self.get_code())
            if argspec.varargs or argspec.varkw:
                self.num_params = -1
            else:
                self.num_params = len(argspec.args)
        return self.num_params

    def get_deterministic(self):
        """Return whether this function is deterministic, that is, always returns the same result
        when called with the same arguments.
        """
        return self.deterministic
    
    def load(self):
        """Register the code of this function into the connection object of the associated SQL store.
        If self.uniquify is True, create a unique name derived from the functions current name.  This
        means the translator needs to access that uniquified name after 'load' was called to create
        a working translated query (see 'translate_call_to_sql').
        """
        code = self.get_code()
        determ = self.get_deterministic()
        if self.uniquify:
            self.uniquify_name()
        self.store.load_user_function(self.get_name(), self.get_num_params(), code, deterministic=determ)

    def uniquify_name(self):
        """Generate a unique name for this function based on its current name.
        """
        self.name = f'{self.get_name()}_{len(self.store.get_user_functions())}'

    def translate_call_to_sql(self, query, expr, state):
        """API method called by the query translator to translate function calls.
        This default method should suffice for most simple user-defined functions.
        """
        args = [query.expression_to_sql(arg, state) for arg in expr.args]
        distinct = expr.distinct and 'DISTINCT ' or ''
        # this will generate a possibly new name, which is why we call this here
        # so we can consider that in the function call translation generated below:
        self.load()
        return f'{self.get_name()}({distinct}{", ".join(args)})'


class AggregateFunction(SqlFunction):
    """Functions that implement aggregation operations.
    """
    def load(self):
        # TO DO: generalize this to call 'self.store.load_aggregate_function()
        #        to properly handle user-defined aggregate functions (see sqlite3 API)
        super().load()

class BuiltinFunction(SqlFunction):
    """Functions that are supported directly by the underlying database.
    These do not need to be defined or loaded.
    """

    def get_code(self):
        """Return the code object of this function which will generally be None
        for builtins, but not raise an error in this case.
        """
        return self.code

    def load(self):
        """Nothing to be done for builtins.
        """
        pass

class BuiltinAggregateFunction(BuiltinFunction, AggregateFunction):
    """Builtin functions that implement aggregation operations.
    """
    pass

class VirtualTableFunction(SqlFunction):
    """Functions that return multiple rows of values per set of inputs and
    operate like a virtual database table and not a scalar function.
    """

    # input parameters and output columns:
    params = ['arg1', 'arg2']
    columns = ['out1', 'out2']

    @staticmethod
    def initialize(vtfun, arg1, arg2='defarg2'):
        """Called by the virtual table function API when a function is called with a set of
        input parameters.  In our adaptation here, this static method becomes the implementation
        of TableFunction.initialize() on the dynamic class we create in self.get_code() below.
        """
        vtfun.arg1 = arg1
        vtfun.arg2 = arg2
        vtfun._result_rows = None

    @staticmethod
    def iterate(vtfun, idx):
        """Called by the virtual table function API when a new set of output values is requested.
        In our adaptation here, this static method becomes the implementation of TableFunction.iterate()
        on the dynamic class we create in self.get_code() below.  This just calls out to
        'vtfun.compute_result_rows()' and should generally not require any specialization on subclasses.
        """
        if vtfun._result_rows is None:
            vtfun._result_rows = vtfun.compute_result_rows()
        return next(vtfun._result_rows)

    @staticmethod
    def compute_result_rows(vtfun):
        """Compute an iterator that produces the result rows for a particular set of inputs.
        Each result row must be a tuple that has values for all output 'columns' specified above.
        Called by the default implementation of vtfun.iterate().
        """
        return iter([(vtfun.arg1, vtfun.arg2)])
    
    def get_num_params(self):
        """Return the number of parameters expected by this function (-1 for variable).
        If not previously specified, infer it from the code object via 'inspect'.
        """
        if self.num_params is None:
            self.num_params = len(self.params)
        return self.num_params

    def get_code(self):
        """Return the code object of this function.  This can return custom adaptations
        relative to a particular function or call context.  For table-valued functions
        the code object has to be represented as a class object subclassing
        playhouse.sqlite_ext.TableFunction, so we create such subclasses dynamically here.
        Custom adaptations can link to additional objects of interest such as vector stores.
        """
        if self.code is None:
            try:
                import playhouse.sqlite_ext as sqlext
            except:
                raise KGTKException(f"you need to 'pip install peewee' to use table-valued functions")
            # multiple-inheritance breaks things, so we inherit the relevant information by hand:
            self.code = type(self.get_name(), (sqlext.TableFunction,), {
                'name': self.get_name(),
                'params': self.params,
                'columns': self.columns,
                'initialize': getattr(type(self), 'initialize'),
                'iterate': getattr(type(self), 'iterate'),
                'compute_result_rows': getattr(type(self), 'compute_result_rows'),
            })
        return self.code
    
    def load(self):
        """Register the code of this function into the connection object of the associated SQL store.
        If self.uniquify is True, create a unique name derived from the functions current name.  This
        means the translator needs to access that uniquified name after 'load' was called to create
        a working translated query (see 'translate_call_to_sql').
        """
        self.uniquify_name()
        code = self.get_code()
        self.store.load_user_function(self.get_name(), self.get_num_params(), code)

"""
# minimal test that uses the dummy definition provided in VirtualTableFunction:
>>> import kgtk.kypher.sqlstore as ss
>>> import kgtk.kypher.functions as fns
>>> sql_store = ss.SqliteStore('/tmp/test.sqlite3.db', loglevel=1)
>>> class TestFun(fns.VirtualTableFunction):
...     pass
... 
>>> TestFun('vtestfun').define()
>>> testfun = fns.SqlFunction.get_function('vtestfun', sql_store)
>>> testfun.load()
>>> testfun.get_name()
'vtestfun_0'
>>> list(sql_store.execute('select * from vtestfun_0(42)'))
[(42, 'defarg2')]
"""

class VirtualGraphFunction(VirtualTableFunction):
    """Functions that return multiple rows of KGTK graph edges per set of inputs.
    These functions should at a minimum take 'node1' as their input and produce
    """



# Top-level definition API:

declare = SqlFunction.declare

def sqlfun(func=None, klass=SqlFunction, name=None, num_params=None, deterministic=False):
    def sqlfun_decorator(func):
        klass(name, code=func, num_params=num_params, deterministic=deterministic).define()
        return func
    if func is None:
        # we are called directly as a function:
        return sqlfun_decorator
    else:
        # we are called as a decorator:
        return sqlfun_decorator(func)

"""
# Examples:

@sqlfun(num_params=1, deterministic=True)
def plus1(x):
    return x + 1

@sqlfun(klass=MathFunction, num_params=1, deterministic=True)
def math_sqrt(x):
    return math.sqrt(x)

CosineSimilarity('kvec_cos_sim', num_params=2, deterministic=True).define()
"""


# Forward-declarations:

declare('kgtk.kypher.funccore',
        'kgtk_regex', 'kgtk_null_to_empty', 'kgtk_empty_to_null', 'pyeval', 'pycall',
        'avg', 'count', 'group_concat', 'max', 'min', 'sum', 'total',
        'cast', 'likelihood', 'concat', 'rowid',
        'kgtk_is_subnode',)

declare('kgtk.kypher.funclit',
        'kgtk_string', 'kgtk_stringify', 'kgtk_unstringify',
        'kgtk_lqstring', 'kgtk_lqstring_text', 'kgtk_lqstring_text_string', 'kgtk_lqstring_lang',
        'kgtk_lqstring_lang_suffix', 'kgtk_lqstring_suffix',
        'kgtk_date', 'kgtk_date_date', 'kgtk_date_time', 'kgtk_date_and_time', 'kgtk_date_year', 'kgtk_date_month', 'kgtk_date_day',
        'kgtk_date_hour', 'kgtk_date_minutes', 'kgtk_date_seconds', 'kgtk_date_zone', 'kgtk_date_zone_string', 'kgtk_date_precision',
        'kgtk_number', 'kgtk_quantity', 'kgtk_quantity_numeral', 'kgtk_quantity_numeral_string', 'kgtk_quantity_number',
        'kgtk_quantity_number_int', 'kgtk_quantity_number_float', 'kgtk_quantity_si_units', 'kgtk_quantity_wd_units',
        'kgtk_quantity_tolerance', 'kgtk_quantity_tolerance_string', 'kgtk_quantity_low_tolerance', 'kgtk_quantity_high_tolerance',
        'kgtk_geo_coords', 'kgtk_geo_coords_lat', 'kgtk_geo_coords_long',
        'kgtk_literal', 'kgtk_symbol', 'kgtk_type',)

declare('kgtk.kypher.funcmath',
        'acos', 'acosh', 'asin', 'asinh', 'atan', 'atan2', 'atanh', 'ceil', 'ceiling', 'cos', 'cosh', 'degrees',
        'exp', 'floor', 'ln', 'log', 'log10', 'log2', 'logb', 'mod', 'pi', 'pow', 'power', 'radians',
        'sin', 'sinh', 'sqrt', 'tan', 'tanh', 'trunc',)

declare('kgtk.kypher.funcvec',
        '_kvec_get_vector', 'kvec_dot', 'kvec_dot_product', 'kvec_cos_sim', 'kvec_cosine_similarity',
        'kvec_topk_cosine_similarity', 'kvec_topk_cos_sim',)



import json
import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry


class TestKGTKCat(unittest.TestCase):
    def setUp(self) -> None:
        self.file_path = 'data/sample_kgtk_edge_file.tsv'
        self.file_path_node = 'data/sample_kgtk_nodes.tsv'
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_cat(self):
        cli_entry("kgtk", "cat", "-i", self.file_path, "-o", f'{self.temp_dir}/cat.tsv', "--verbose", "--show-option")

        df_r = pd.read_csv(f'{self.temp_dir}/cat.tsv', sep='\t')

        df = pd.read_csv(f'{self.file_path}', sep='\t')
        self.assertEqual(len(df), len(df_r))
        self.assertEqual(list(df.columns), list(df_r.columns))

        f = open(self.file_path)
        lines = f.readlines()
        f.close()
        for i, row in df_r.iterrows():
            self.assertEqual(row["id"], lines[i + 1].split('\t')[0])

    def test_kgtk_cat_output_csv(self):
        cli_entry("kgtk", "cat", "-i", self.file_path, "-o", f'{self.temp_dir}/cat.csv', "--output-format", "csv")

        df_r = pd.read_csv(f'{self.temp_dir}/cat.csv')

        df = pd.read_csv(f'{self.file_path}', sep='\t')
        self.assertEqual(len(df), len(df_r))
        self.assertEqual(list(df.columns), list(df_r.columns))

        f = open(self.file_path)
        lines = f.readlines()
        f.close()
        for i, row in df_r.iterrows():
            self.assertEqual(row["id"], lines[i + 1].split('\t')[0])

    def test_kgtk_cat_output_json_map(self):
        cli_entry("kgtk", "cat", "-i", self.file_path, "-o", f'{self.temp_dir}/cat.json', "--output-format", "json-map")

        f = open(f'{self.temp_dir}/cat.json')
        obj = json.load(f)
        f.close()
        self.assertEqual(len(obj), 287)
        for x in obj:
            self.assertTrue('id' in x)
            self.assertTrue('node1' in x)
            self.assertTrue('label' in x)
            self.assertTrue('node2' in x)
            self.assertTrue('rank' in x)

    def test_kgtk_cat_output_json_line(self):
        cli_entry("kgtk", "cat", "-i", self.file_path, "-o", f'{self.temp_dir}/cat.jl', "--output-format", "jsonl")

        f = open(f'{self.temp_dir}/cat.jl')
        lines = f.readlines()
        f.close()
        self.assertEqual(len(lines), 288)
        for line in lines:
            self.assertEqual(len(json.loads(line)), 5)

    def test_kgtk_cat_output_json_line_map(self):
        cli_entry("kgtk", "cat", "-i", self.file_path, "-o", f'{self.temp_dir}/cat.jl', "--output-format", "jsonl-map")

        f = open(f'{self.temp_dir}/cat.jl')
        lines = f.readlines()
        f.close()
        self.assertEqual(len(lines), 287)
        for line in lines:
            x = json.loads(line)
            self.assertTrue('id' in x)
            self.assertTrue('node1' in x)
            self.assertTrue('label' in x)
            self.assertTrue('node2' in x)
            self.assertTrue('rank' in x)

    def test_kgtk_cat_edge_node_file(self):
        rc = cli_entry("kgtk", "cat", "-i", self.file_path, self.file_path_node, "-o", f'{self.temp_dir}/cat.tsv')
        # non zero exit code
        self.assertEqual(rc, 1)

    def test_kgtk_cat_two_files(self):
        f1_path = 'data/sample_kgtk_edge_Q47158.tsv'
        f2_path = 'data/sample_kgtk_edge_file_with_id.tsv'
        cli_entry("kgtk", "cat", "-i", f1_path, f2_path, "-o", f'{self.temp_dir}/cat.tsv')
        df = pd.read_csv(f'{self.temp_dir}/cat.tsv', sep='\t')
        self.assertEqual(len(df), 6)



"""Add label columns for values in the node1, label, and node2 fields.

The input rows are saved in memory, as well as the value-to-label mapping.
This will impose a limit on the size of the input files that can be processed.

TODO: Optionally save the input rows in an external disk file?

TODO: Optionally reread the input stream insted of saving the input rows?
      With special provision for copying standard input to an external disk file.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

LIFT_COMMAND: str = 'lift'
ADD_LABELS_COMMAND: str = 'add-labels'

def parser():
    return {
        'aliases': [ ADD_LABELS_COMMAND ],
        'help': 'Lift labels from a KGTK file.',
        'description': 'Lift labels for a KGTK file. ' +
        'If called as "kgtk lift", for each of the items in the (node1, label, node2) columns, look for matching label records. ' +
        'If called as "kgtk add-labels", look for matching label records for all input columns. ' +
        'If found, lift the label values into additional columns in the current record. ' +
        'Label records are removed from the output unless --remove-label-records=False. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert lift --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.exceptions import KGTKException
    from kgtk.lift.kgtklift import KgtkLift
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert
    _command: str = parsed_shared_args._command

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_input_file(who="A KGTK file with label records",
                          dest="label_file",
                          options=["--label-file"],
                          optional=True)

    parser.add_output_file(who="A KGTK output file that will contain only unmodified rows." +
                           " This file will have the same columns as the input file.",
                           dest="unmodified_row_file",
                           options=["--unmodified-row-output-file"],
                           metavar="UNMODIFIED_ROW_OUTPUT_FILE",
                           optional=True)

    parser.add_output_file(who="A KGTK output file that will contain matched label edges." +
                           " This file will have the same columns as the source of the labels, either the input file or the label file.",
                           dest="matched_label_file",
                           options=["--matched-label-output-file"],
                           metavar="MATCHED_LABEL_OUTPUT_FILE",
                           optional=True)

    parser.add_output_file(who="A KGTK output file that will contain unmatched label edges." +
                           " This file will have the same columns as the source of the labels, either the input file or the label file.",
                           dest="unmatched_label_file",
                           options=["--unmatched-label-output-file"],
                           metavar="UNMATCHED_LABEL_OUTPUT_FILE",
                           optional=True)

    suppress_empty_columns_default: bool
    prefilter_labels_default: bool
    use_label_enver_default: bool
    lift_all_columns_default: bool
    require_label_file_default: bool
    force_input_mode_none_default: bool
    if _command == LIFT_COMMAND:
        parser.add_argument(      "--columns-to-lift", dest="input_lifting_column_names",
                                  help=h("The columns for which matching labels are to be lifted. " +
                                         "The default is [node1, label, node2] or their aliases."), nargs='*')

        parser.add_argument(      "--columns-to-write", dest="output_lifted_column_names",
                                  help="The columns into which to store the lifted values. " +
                                  "The default is [node1;label, label;label, node2;label] or their aliases.", nargs='*')

        suppress_empty_columns_default = False
        prefilter_labels_default = False
        use_label_envar_default = False
        lift_all_columns_default = False
        require_label_file_default = False
        force_input_mode_none_default = False
       
    elif _command == ADD_LABELS_COMMAND:
        parser.add_argument(      "--columns-to-lift", dest="input_lifting_column_names",
                                  help=h("The columns for which matching labels are to be lifted. " +
                                         "The default is all columns except id or its alias."), nargs='*')

        parser.add_argument(      "--columns-to-write", dest="output_lifted_column_names",
                                  help="The columns into which to store the lifted values. " +
                                  "The default is [node1;label, label;label, node2;label, ...] or their aliases.", nargs='*')

        suppress_empty_columns_default = True
        prefilter_labels_default = True
        use_label_envar_default = True
        lift_all_columns_default = True
        require_label_file_default = True
        force_input_mode_none_default = True

    else:
        raise KGTKException("Unknown command %s" % repr(_command))

    parser.add_argument(      "--input-select-column", "--input-label-column", dest="input_select_column_name",
                              help=h("If input record selection is enabled by --input-select-value, " +
                              "the name of a column that determines which records received lifted values. " +
                              "The default is the 'label' column or its alias."), default=None)

    parser.add_argument(      "--input-select-value", "--input-label-value", "--target-label-value", dest="input_select_column_value",
                              help=h("The value in the input select column that identifies a record to receive lifted values. " +
                              "The default is not to perform input record selection, " +
                              "and all input records except label records may receive lifted values. "),
                              default=None)

    parser.add_argument(      "--lift-suffix", dest="output_lifted_column_suffix",
                              help=h("The suffix used for newly created output columns. (default=%(default)s)."),
                              default=KgtkLift.DEFAULT_OUTPUT_LIFTED_COLUMN_SUFFIX)

    parser.add_argument(      "--default-value", dest="default_value",
                              help="The value to use if a lifted label is not found. (default=%(default)s)", default="")

    parser.add_argument(      "--update-select-value", "--target-new-label-value", dest="output_select_column_value",
                              help=h("A new value for the select (label) column for records that received lifted values. " +
                              "The default is not to update the select(label) column."), default=None)
    

    parser.add_argument(      "--label-select-column", "--label-name", dest="label_select_column_name",
                              help=h("The name of the column that contains a special value that identifies label records. " +
                              "The default is 'label' or its alias."), default=None)

    parser.add_argument("-p", "--label-select-value", "--label-value", "--property", dest="label_select_column_value",
                              help=h("The special value in the label select column that identifies a label record. " +
                              "(default=%(default)s)."), default=KgtkLift.DEFAULT_LABEL_SELECT_COLUMN_VALUE)
    
    parser.add_argument(      "--label-match-column", "--node1-name", dest="label_match_column_name",
                              help=h("The name of the column in the label records that contains the value " +
                              "that matches the value in a column being lifted in the input records. " +
                              "The default is 'node1' or its alias."), default=None)

    parser.add_argument(      "--label-value-column", "--node2-name", "--lift-from", dest="label_value_column_name",
                              help=h("The name of the column in the label record that contains the value " +
                              "to be lifted into the input record that is receiving lifted values. " +
                              "The default is 'node2' or its alias."), default=None)

    parser.add_argument(      "--remove-label-records", dest="remove_label_records",
                              help=h("If true, remove label records from the output. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--sort-lifted-labels", dest="sort_lifted_labels",
                              help=h("If true, sort lifted labels with lists. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-duplicate-labels", dest="suppress_duplicate_labels",
                              help=h("If true, suppress duplicate values in lists in lifted labels (implies sorting). (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-empty-columns", dest="suppress_empty_columns",
                              help="If true, do not create new columns that would be empty. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=suppress_empty_columns_default)


    parser.add_argument(      "--ok-if-no-labels", dest="ok_if_no_labels",
                              help="If true, do not abort if no labels were found. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--prefilter-labels", dest="prefilter_labels",
                              help="If true, read the input file before reading the label file. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=prefilter_labels_default)

    parser.add_argument(      "--input-file-is-presorted", dest="input_is_presorted",
                              help="If true, the input file is presorted on the column for which values are to be lifted. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--label-file-is-presorted", dest="labels_are_presorted",
                              help="If true, the label file is presorted on the node1 column. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--clear-before-lift", dest="clear_before_lift",
                              help="If true, set columns to write to the default value before lifting. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--overwrite", dest="overwrite",
                              help="If true, overwrite non-default values in the columns to write. " +
                              "If false, do not overwrite non-default values in the columns to write. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--output-only-modified-rows", dest="output_only_modified_rows",
                              help="If true, output only modified edges to the primary output stream. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--languages", dest="languages", type=str, nargs="*", metavar="LANGUAGE",
                              help="Lift only labels with a matching language qualifier. " +
                              "ANY means any language qualifier. NONE means no language qualifier. (default=ANY NONE)")

    parser.add_argument(      "--prioritize", dest="prioritize",
                              help="If true and filtering labels by language, pick only the label matching " +
                              "the language that appears before other matches in the language list. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=use_label_envar_default)

    parser.add_argument(      "--use-label-envar", dest="use_label_envar",
                              help="If true, use the KGTK_LABEL_FILE envar for the label file if no --label-file. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=use_label_envar_default)

    parser.add_argument(      "--lift-all-columns", dest="lift_all_columns",
                              help=h("If true, lift all columns. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=lift_all_columns_default)
        
    parser.add_argument(      "--require-label-file", dest="require_label_file",
                              help=h("If true, a label file must be required explicitly or through an envar. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=require_label_file_default)
        
    parser.add_argument(      "--force-input-mode-none", dest="force_input_mode_none",
                              help=h("If true, force the input file to mode NONE. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=force_input_mode_none_default)
        
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    # TODO: seperate reader_options for the label file.
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="label", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        label_file: KGTKFiles,
        unmodified_row_file: KGTKFiles,
        matched_label_file: KGTKFiles,
        unmatched_label_file: KGTKFiles,

        input_select_column_name: typing.Optional[str],
        input_select_column_value: typing.Optional[str],
        input_lifting_column_names: typing.List[str],

        output_lifted_column_names: typing.List[str],
        output_lifted_column_suffix: str,
        output_select_column_value: str,

        label_select_column_name: typing.Optional[str],
        label_select_column_value: str,
        label_match_column_name: typing.Optional[str],
        label_value_column_name: typing.Optional[str],

        default_value: str,

        remove_label_records: bool = False,
        sort_lifted_labels: bool = True,
        suppress_duplicate_labels: bool = True,
        suppress_empty_columns: bool = False,
        ok_if_no_labels: bool = False,
        prefilter_labels: bool = False,
        input_is_presorted: bool = False,
        labels_are_presorted: bool = False,

        clear_before_lift: bool = False,
        overwrite: bool = False,

        output_only_modified_rows: bool = False,

        languages: typing.Optional[typing.List[str]] = None,
        prioritize: bool = False,

        use_label_envar: bool = False,
        lift_all_columns: bool = False,
        require_label_file: bool = False,
        force_input_mode_none: bool = False,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    import os
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.lift.kgtklift import KgtkLift
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    label_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_input_file(label_file, who="KGTK label file")
    unmodified_row_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(unmodified_row_file, who="KGTK unmodified row output file")
    matched_label_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(matched_label_file, who="KGTK matched label output file")
    unmatched_label_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(unmatched_label_file, who="KGTK unmatched label output file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
    label_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="label", fallback=True)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        if label_kgtk_file is not None:
            print("--label-file=%s" % label_kgtk_file, file=error_file, flush=True)
        if unmodified_row_kgtk_file is not None:
            print("--unmodified-row-output-file=%s" % unmodified_row_kgtk_file, file=error_file, flush=True)
        if matched_label_kgtk_file is not None:
            print("--matched-label-output-file=%s" % matched_label_kgtk_file, file=error_file, flush=True)
        if unmatched_label_kgtk_file is not None:
            print("--unmatched-label-output-file=%s" % unmatched_label_kgtk_file, file=error_file, flush=True)

        if input_select_column_name is not None:
            print("--input-select-column=%s" % input_select_column_name, file=error_file, flush=True)
        if input_select_column_value is not None:
            print("--input-select-value=%s" % input_select_column_value, file=error_file, flush=True)
        if input_lifting_column_names is not None and len(input_lifting_column_names) > 0:
            print("--columns-to-lift %s" % " ".join(input_lifting_column_names), file=error_file, flush=True)
        if output_lifted_column_names is not None and len(output_lifted_column_names) > 0:
            print("--columns-to-write %s" % " ".join(output_lifted_column_names), file=error_file, flush=True)

        print("--lift-suffix=%s" % output_lifted_column_suffix, file=error_file, flush=True)
        if output_select_column_value is not None:
            print("--update-select-value=%s" % output_select_column_value, file=error_file, flush=True)


        if label_select_column_name is not None:
            print("--label-select-column=%s" % label_select_column_name, file=error_file, flush=True)
        print("--label-select-value=%s" % label_select_column_value, file=error_file, flush=True)
        if label_match_column_name is not None:
            print("--label-match-column=%s" % label_match_column_name, file=error_file, flush=True)
        if label_value_column_name is not None:
            print("--label-value-column=%s" % label_value_column_name, file=error_file, flush=True)

        print("--default-value=%s" % repr(default_value), file=error_file, flush=True)
        print("--remove-label-records=%s" % repr(remove_label_records), file=error_file, flush=True)
        print("--sort-lifted-labels=%s" % repr(sort_lifted_labels), file=error_file, flush=True)
        print("--suppress-duplicate-labels=%s" % repr(suppress_duplicate_labels), file=error_file, flush=True)
        print("--suppress-empty-columns=%s" % repr(suppress_empty_columns), file=error_file, flush=True)
        print("--ok-if-no-labels=%s" % repr(ok_if_no_labels), file=error_file, flush=True)
        print("--prefilter-labels=%s" % repr(prefilter_labels), file=error_file, flush=True)
        print("--input-file-is-presorted=%s" % repr(input_is_presorted), file=error_file, flush=True)
        print("--label-file-is-presorted=%s" % repr(labels_are_presorted), file=error_file, flush=True)
        print("--clear-before-lift=%s" % repr(clear_before_lift), file=error_file, flush=True)
        print("--overwrite=%s" % repr(overwrite), file=error_file, flush=True)
        print("--output-only-modified-rows=%s" % repr(output_only_modified_rows), file=error_file, flush=True)

        if languages is not None:
            print("--languages %s" % " ".join(repr(l) for l in languages), file=error_file, flush=True)
        print("--prioritize=%s" % repr(prioritize), file=error_file, flush=True)
        
        print("--use-label-envar=%s" % repr(use_label_envar), file=error_file, flush=True)
        print("--lift-all-columns=%s" % repr(lift_all_columns), file=error_file, flush=True)
        print("--require-label-files=%s" % repr(require_label_file), file=error_file, flush=True)
        print("--force-input-mode-none=%s" % repr(force_input_mode_none), file=error_file, flush=True)
        input_reader_options.show(out=error_file, who="input")
        label_reader_options.show(out=error_file, who="label")
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Should the following functionality be moved to KgtkLift?
    if label_kgtk_file is None and use_label_envar:
        label_file_envar: str = 'KGTK_LABEL_FILE' # TODO: Move this to a common file.
        label_file_envar_value: typing.Optional[str] = os.getenv(label_file_envar)
        if label_file_envar_value is not None:
            label_kgtk_file = Path(label_file_envar_value)
            if verbose:
                print("Using label file %s from envar %s" % (repr(label_file_envar_value), repr(label_file_envar)), file=error_file, flush=True)

    if require_label_file and label_kgtk_file is None:
        raise KGTKException("A label file must be specified using --label-file or KGTK_LABEL_FILE")

    try:
        kl: KgtkLift = KgtkLift(
            input_file_path=input_kgtk_file,
            label_file_path=label_kgtk_file,
            output_file_path=output_kgtk_file,
            unmodified_row_file_path=unmodified_row_kgtk_file,
            matched_label_file_path=matched_label_kgtk_file,
            unmatched_label_file_path=unmatched_label_kgtk_file,

            input_select_column_name=input_select_column_name,
            input_select_column_value=input_select_column_value,
            input_lifting_column_names=input_lifting_column_names,

            output_lifted_column_suffix=output_lifted_column_suffix,
            output_select_column_value=output_select_column_value,
            output_lifted_column_names=output_lifted_column_names,

            label_select_column_name=label_select_column_name,
            label_select_column_value=label_select_column_value,
            label_match_column_name=label_match_column_name,
            label_value_column_name=label_value_column_name,

            default_value=default_value,

            remove_label_records=remove_label_records,
            sort_lifted_labels=sort_lifted_labels,
            suppress_duplicate_labels=suppress_duplicate_labels,
            suppress_empty_columns=suppress_empty_columns,
            ok_if_no_labels=ok_if_no_labels,
            prefilter_labels=prefilter_labels,
            input_is_presorted=input_is_presorted,
            labels_are_presorted=labels_are_presorted,

            clear_before_lift=clear_before_lift,
            overwrite=overwrite,

            output_only_modified_rows=output_only_modified_rows,

            languages=languages,
            prioritize=prioritize,

            lift_all_columns=lift_all_columns,
            force_input_mode_none=force_input_mode_none,

            input_reader_options=input_reader_options,
            label_reader_options=label_reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        kl.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""This utility is analogous to the POSIX 'tail' command.

For "-n N",  pass just the last N data edges of a KGTK input file to the KGTK output file.

The POSIX 'tail' command's notion of '-n +N' is not supported.

The header record, cotaining the column names, is always passed and is not
included in N.

Multiplier suffixes are not supported.

--mode=NONE is default.

TODO: Need KgtkWriterOptions

"""

from argparse import Namespace, SUPPRESS

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Pass the tail (last records) of a KGTK file.',
        'description': 'This utility is analogous to the POSIX "head" command. ' +
        '\n\nFor "-n N", pass just the last N data edges of a KGTK input file to the KGTK output file. ' +
        '\n\n"-n +N" does not have the special meaning it has in the POSIC "tail" command. ' +
        '\n\nThe header record, cotaining the column names, is always passed and is not included in N. ' +
        '\n\nMultiplier suffixes are not supported. ' +
        '\n\nUse this command to filter the output of any KGTK command: ' +
        '\n\nkgtk xxx / tail -n 20 ' +
        '\n\nUse it to limit the records in a file: ' +
        '\n\nkgtk tail -i file.tsv -o file.html' +
        '\n\nThis command defaults to --mode=NONE so it will work with TSV files that do not follow KGTK column naming conventions.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert html --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file()
    parser.add_output_file()

    parser.add_argument("-n", "--edges", dest="edge_limit", type=int, default=10,
                        help="The number of records to pass (default=%(default)d).")
                        
    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=kgtk)"), type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        edge_limit: int,
        output_format: str,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from collections import deque
    from pathlib import Path
    import sys
    import typing
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, mode=KgtkReaderMode.NONE)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        print("--edges=%s" % str(edge_limit), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kr: KgtkReader = KgtkReader.open(input_file_path,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )

        output_mode: KgtkWriter.Mode = KgtkWriter.Mode.NONE
        if kr.is_edge_file:
            output_mode = KgtkWriter.Mode.EDGE
            if verbose:
                print("Opening the output edge file: %s" % str(output_file_path), file=error_file, flush=True)

        elif kr.is_node_file:
            output_mode = KgtkWriter.Mode.NODE
            if verbose:
               print("Opening the output node file: %s" % str(output_file_path), file=error_file, flush=True)

        else:
            if verbose:
                print("Opening the output file: %s" % str(output_file_path), file=error_file, flush=True)

        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         output_file_path,
                                         use_mgzip=reader_options.use_mgzip, # Hack!
                                         mgzip_threads=reader_options.mgzip_threads, # Hack!
                                         gzip_in_parallel=False,
                                         mode=output_mode,
                                         output_format=output_format,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        edge_count: int = 0
        row: typing.List[str]
        edge_buffer: deque = deque()
        for row in kr:
            edge_buffer.append(row)
            if len(edge_buffer) > edge_limit:
                edge_buffer.popleft()

        while len(edge_buffer) > 0:
            edge_count += 1
            kw.write(edge_buffer.popleft())

        kw.close()

        if verbose:
            print("Copied %d edges." % edge_count, file=error_file, flush=True)

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




from kgtk.knowledge_graph.subject import Subject
from kgtk.knowledge_graph.node import URI, BNode
from kgtk.knowledge_graph.namespacemanager import NamespaceManager
from functools import lru_cache
import rdflib


class Graph(object):
    def __init__(self, graph=None, identifier=None):
        if graph:
            self._g = graph
        else:
            self._g = rdflib.Graph(identifier=identifier)
        self._ns = NamespaceManager(self._g)

    def bind(self, prefix, namespace, override=True, replace=False):
        self._ns.bind(prefix, namespace, override, replace)

    def add_subject(self, subjects, context=None):
        if not context:
            context = set()

        for t in subjects:
            s, p, o = t
            if isinstance(o, Subject) and o not in context:
                context.add(o)
                self.add_subject(o, context)
                o = o.subject

            # convert triple to RDFLib recognizable format
            triple = self._convert_triple_rdflib((s, p, o))
            self._g.add(triple)

    def add_triple(self, s, p, o):
        t = Subject(s)
        t.add_property(p, o)
        self.add_subject(t)

    def parse(self, content, format='turtle'):
        self._g.parse(data=content, format=format)

    def serialize(self, format='ttl', namespace_manager=None, **kwargs):
        # may need some way to serialize ttl, json-ld
        if format.lower() in ('ttl', 'turtle'):
            b_string = self._g.serialize(format=format, namespace_manager=namespace_manager, **kwargs)
        elif format.lower() == 'json-ld':
            b_string = self._g.serialize(format=format, contexts=namespace_manager, **kwargs)
        else:
            b_string = self._g.serialize(format=format, **kwargs)
        if isinstance(b_string, bytes):
            b_string = b_string.decode('UTF-8')
        return b_string

    @lru_cache()
    def _resolve_uri(self, uri: URI) -> rdflib.URIRef:
        """
        Convert a URI object into a RDFLib URIRef, including resolve its context

        :param uri: URI
        :return: rdflib.URIRef
        """
        return self._ns.parse_uri(uri.value)

    def _is_rdf_type(self, uri: URI) -> bool:
        if not isinstance(uri, URI):
            return False
        return self._resolve_uri(uri) == rdflib.RDF.type

    def _convert_triple_rdflib(self, triple):
        """
        Convert a Node triple into RDFLib triple
        """
        s, p, o = triple
        sub = self._resolve_uri(s) if isinstance(s, URI) else rdflib.BNode(s.value)
        pred = self._resolve_uri(p)
        if isinstance(o, URI):
            obj = self._resolve_uri(o)
        elif isinstance(o, Subject):
            if isinstance(o.subject, URI):
                obj = self._resolve_uri(o.subject)
            else:
                obj = rdflib.BNode(o.subject.value)
        elif isinstance(o, BNode):
            obj = rdflib.BNode(o.value)
        else:
            obj = rdflib.Literal(o.value, o.lang, o.raw_type)
        return sub, pred, obj



from typing import Dict
from kgtk.knowledge_graph.knowledge_graph import KnowledgeGraph


class Document(object):

    def __init__(self, _document: Dict, kg_schema, doc_id=None) -> None:
        self._document = _document
        if doc_id:
            self._document["doc_id"] = doc_id

        self.kg = KnowledgeGraph(kg_schema, self)



"""
Import WordNet to KGTK.

"""

import sys
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import WordNet into KGTK.' 
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    parser.add_output_file()

def run(output_file: KGTKFiles):

    # import modules locally
    import sys # type: ignore
    from kgtk.exceptions import kgtk_exception_auto_handler
    import json
    import nltk
    nltk.download("wordnet")
    nltk.download('omw-1.4')
    from nltk.corpus import wordnet as wn
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkwriter import KgtkWriter

    def obtain_wordnet_lemmas(syn):
        lemmas=[]
        for lemma in syn.lemma_names():
            lemmas.append(KgtkFormat.stringify(lemma.replace('_', ' ')))
        return lemmas

    def obtain_hypernyms(syn):
        hyps=[]
        for hypernym in syn.hypernyms():
            hyps.append(hypernym.name())
        return hyps

    def obtain_member_holonyms(syn):
        hols=[]
        for hol in syn.member_holonyms():
            hols.append(hol.name())
        return hols

    def obtain_part_holonyms(syn):
        hols=[]
        for hol in syn.part_holonyms():
            hols.append(hol.name())
        return hols

    def obtain_substance_meronyms(syn):
        hols=[]
        for hol in syn.substance_meronyms():
            hols.append(hol.name())
        return hols


    def get_wn_data():
        syns=list(wn.all_synsets())
        all_labels={}
        all_hyps={}
        all_members={}
        all_parts={}
        all_subs={}
        for syn in syns:
            syn_name=syn.name()

            lemmas=obtain_wordnet_lemmas(syn)
            all_labels[syn_name]='|'.join(lemmas)

            hypernyms=obtain_hypernyms(syn)
            if len(hypernyms):
                all_hyps[syn_name]=hypernyms
            
            member_holonyms=obtain_member_holonyms(syn)
            if len(member_holonyms):
                all_members[syn_name]=member_holonyms

            part_holonyms=obtain_part_holonyms(syn)
            if len(part_holonyms):
                all_parts[syn_name]=part_holonyms

            substance_meronyms=obtain_substance_meronyms(syn)
            if len(substance_meronyms):
                all_subs[syn_name]=substance_meronyms

        return all_labels, all_hyps, all_members, all_parts, all_subs

    def create_edges(data, labels, rel, rel_label):
        all_rows=[]
        source=KgtkFormat.stringify('WN')
        for node1, v in data.items():
            for node2 in v:
                node1_preflabel=labels[node1].split('|')[0]
                node2_preflabel=labels[node2].split('|')[0]
                a_row=['wn:' + node1, rel, 'wn:' + node2, labels[node1], labels[node2], rel_label, "", source, '']
                all_rows.append(a_row)
        return all_rows

    try:
        out_columns=['node1', 'relation', 'node2', 'node1;label', 'node2;label','relation;label', 'relation;dimension', 'source', 'sentence']

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
        ew: KgtkWriter = KgtkWriter.open(out_columns,
                                         output_kgtk_file,
                                         #mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         #verbose=self.verbose,
                                         #very_verbose=self.very_verbose
                                         )

        all_labels, all_hyps, all_members, all_parts, all_subs=get_wn_data()
        hyp_edges=create_edges(all_hyps, all_labels, '/r/IsA', KgtkFormat.stringify('is a'))
        member_edges=create_edges(all_members, all_labels, '/r/PartOf', KgtkFormat.stringify('is a part of'))
        part_edges=create_edges(all_parts, all_labels, '/r/PartOf', KgtkFormat.stringify('is a part of'))
        sub_edges=create_edges(all_subs, all_labels, '/r/MadeOf', KgtkFormat.stringify('is made of'))
        all_edges=hyp_edges+member_edges+part_edges+sub_edges

        for edge in all_edges:
            ew.write(edge)

        # Clean up.
        ew.close()

    except Exception as e:
            kgtk_exception_auto_handler(e)



"""Import ntriples into KGTK format.

The W3C document on RDF 1.1 N-Triples is https://www.w3.org/TR/n-triples/.

The W3C document defines optional
comments that may appear after the full stop (".") at the end of a triple. COmments
begin with "#" (after optional white space) and continue ot the end of line or end of file.

At present, `kgtk import-ntriples` strips any comments found, with a summary count at the end
of processing. A more complex option would be to convert N-Triples comments into KGTK comments.


Issue #390: KGTK Needs More General Language Tags

According to the Wikipedia article on N-Triples, N-triples use RFC 3066
language tags. According to the Wikipedia article on IETF language tags,
RFC 3066 has been replaced by RFC 4646 and RFC 5646, with increasingly
more general structure.

According to the W3C document on RDF 1.1 N-triples, language tags are
defined by BCP 47, which references RFC 5646. The Wikipedia article appears
to be obsolete.

To support importing/exporting N-Triples properly, the KGTK specification and
implementation may need some extensions.


"""
from argparse import ArgumentParser, Namespace
import ast
import attr
import csv
from pathlib import Path
import re
import shortuuid # type: ignore
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue, KgtkValueFields
from kgtk.value.kgtkvalueoptions import KgtkValueOptions, DEFAULT_KGTK_VALUE_OPTIONS

@attr.s(slots=True, frozen=False)
class KgtkNtriples(KgtkFormat):
    # Class attributes:
    DEFAULT_PREFIX_EXPANSION: str = "prefix_expansion"
    DEFAULT_STRUCTURED_VALUE: str = "kgtk:structured_value"
    DEFAULT_STRUCTURED_URI: str = "kgtk:structured_uri"
    DEFAULT_NAMESPACE_ID_PREFIX: str = "n"
    DEFAULT_NAMESPACE_ID_COUNTER: int = 1
    DEFAULT_NAMESPACE_ID_ZFILL: int = 0
    DEFAULT_NAMESPACE_ID_USE_UUID: bool = False
    DEFAULT_OUTPUT_ONLY_USED_NAMESPACES: bool = True
    DEFAULT_BUILD_NEW_NAMESPACES: bool = True
    DEFAULT_NEWNODE_PREFIX: str = "kgtk:node"
    DEFAULT_NEWNODE_COUNTER: int = 1
    DEFAULT_NEWNODE_ZFILL: int = 0
    DEFAULT_NEWNODE_USE_UUID: bool = False
    DEFAULT_LOCAL_NAMESPACE_PREFIX: str = "X"
    DEFAULT_LOCAL_NAMESPACE_USE_UUID: bool = True
    DEFAULT_ALLOW_LAX_URI: bool = True
    DEFAULT_BUILD_ID: bool = False
    DEFAULT_VALIDATE: bool = False
    DEFAULT_ALLOW_UNKNOWN_DATATYPE_IRIS: bool = True
    DEFAULT_ALLOW_TURTLE_QUOTES: bool = False
    DEFAULT_ALLOW_LANG_STRING_DATATYPE: bool = False
    DEFAULT_SUMMARY: bool = False
    LANG_STRING_TAG_NONE: str = "-"
    DEFAULT_LANG_STRING_TAG: str = LANG_STRING_TAG_NONE
    DEFAULT_BUILD_DATATYPE_COLUMN: bool = False
    DEFAULT_DATATYPE_COLUMN_NAME: str = "datatype"
    DEFAULT_WRITE_NAMESPACES: bool = True

    COLUMN_NAMES: typing.List[str] = [KgtkFormat.NODE1, KgtkFormat.LABEL, KgtkFormat.NODE2]
    
    # The following lexical analysis is based on:
    # https://docs.python.org/3/reference/lexical_analysis.html

    # The long integer suffix was part of Python 2.  It was dropped in Python 3.
    long_suffix_pat: str = r'[lL]'

    plus_or_minus_pat: str = r'[-+]'

    # Integer literals.
    #
    # Decimal integers, allowing leading zeros.
    digit_pat: str = r'[0-9]'
    decinteger_pat: str = r'(?:{digit}(?:_?{digit})*{long_suffix}?)'.format(digit=digit_pat,
                                                                            long_suffix=long_suffix_pat)
    bindigit_pat: str = r'[01]'
    bininteger_pat: str = r'(?:0[bB](":_?{bindigit})+{long_suffix})'.format(bindigit=bindigit_pat,
                                                                            long_suffix=long_suffix_pat)
    octdigit_pat: str = r'[0-7]'
    octinteger_pat: str = r'(?:0[oO](":_?{octdigit})+{long_suffix})'.format(octdigit=octdigit_pat,
                                                                            long_suffix=long_suffix_pat)
    hexdigit_pat: str = r'[0-7a-fA-F]'
    hexinteger_pat: str = r'(?:0[xX](":_?{hexdigit})+{long_suffix})'.format(hexdigit=hexdigit_pat,
                                                                            long_suffix=long_suffix_pat)
     
    integer_pat: str = r'(?:{decinteger}|{bininteger}|{octinteger}|{hexinteger})'.format(decinteger=decinteger_pat,
                                                                                         bininteger=bininteger_pat,
                                                                                         octinteger=octinteger_pat,
                                                                                         hexinteger=hexinteger_pat)

    # Floating point literals.
    digitpart_pat: str = r'(?:{digit}(?:_?{digit})*)'.format(digit=digit_pat)
    fraction_pat: str = r'(?:\.{digitpart})'.format(digitpart=digitpart_pat)
    pointfloat_pat: str = r'(?:{digitpart}?{fraction})|(?:{digitpart}\.)'.format(digitpart=digitpart_pat,
                                                                                 fraction=fraction_pat)
    exponent_pat: str = r'(?:[eE]{plus_or_minus}?{digitpart})'.format(plus_or_minus=plus_or_minus_pat,
                                                                      digitpart=digitpart_pat)
    exponentfloat_pat: str = r'(?:{digitpart}|{pointfloat}){exponent}'.format(digitpart=digitpart_pat,
                                                                              pointfloat=pointfloat_pat,
                                                                              exponent=exponent_pat)
    floatnumber_pat: str = r'(?:{pointfloat}|{exponentfloat})'.format(pointfloat=pointfloat_pat,
                                                                      exponentfloat=exponentfloat_pat)

    # Imaginary literals.
    imagnumber_pat: str = r'(?:{floatnumber}|{digitpart})[jJ]'.format(floatnumber=floatnumber_pat,
                                                                      digitpart=digitpart_pat)

    # Numeric literals.
    numeric_pat: str = r'(?:{plus_or_minus}?(?:{integer}|{floatnumber}|{imagnumber}))'.format(plus_or_minus=plus_or_minus_pat,
                                                                                              integer=integer_pat,
                                                                                              floatnumber=floatnumber_pat,
                                                                                              imagnumber=imagnumber_pat)

    # A URI must begin with a scheme per RFC 3986.
    #
    # We don't use this because ISI was inserting some invalid URIs that did
    # not include a scheme, but kept the definition for reference.
    SCHEME_PAT: str = r'^(?:[a-zA_Z][-.+0-9a-zA_Z]*://)'

    # The URI is delimited by matching angle brackets.  Angle brackets
    # cannot appear in a valid URI per RFC 3986.
    URI_PAT: str = r'(?:<[^>]+>)'

    # This is a guess about what may be in a blank node. It's entirely
    # possible that other characters, sych as hyphen, might be allowed.
    BLANK_NODE_PAT: str = r'(?:_:[0-9a-zA-Z_]+)'

    # Double quoted strings with backslash escapes.
    STRING_PAT: str = r'(?:"(?:[^\\]|(?:\\.))*")'

    # Single quoted strings with backslash escapes are allowed in RDF Turtle format.
    TURTLE_STRING_PAT: str = r"(?:'(?:[^\\]|(?:\\.))*')"

    # Language tags are sequences of alphanumerics separated by dashes.
    LANGUAGE_TAG_PAT: str = r'(?:[0-9a-zA-Z]+(?:-[0-9a-zA-Z]+)*)'

    STRUCTURED_VALUE_PAT: str = r'(?:(?:{string}|{turtle_string})(?:(?:@{tag})|(?:\^\^{uri}))?)'.format(string=STRING_PAT,
                                                                                                        turtle_string=TURTLE_STRING_PAT,
                                                                                                        tag=LANGUAGE_TAG_PAT,
                                                                                                        uri=URI_PAT)
    FIELD_PAT: str = r'(?:{uri}|{blank_node}|{structured_value}|{numeric})'.format(uri=URI_PAT,
                                                                                   blank_node=BLANK_NODE_PAT,
                                                                                   structured_value=STRUCTURED_VALUE_PAT,
                                                                                   numeric=numeric_pat,
    )
    COMMENT_PAT: str = r'(?:#.*)?'
    ROW_PAT: str = r'\s*(?P<node1>{field})\s+(?P<label>{field})\s+(?P<node2>{field})\s+\.\s*(?P<comment>{comment})'.format(field=FIELD_PAT, comment=COMMENT_PAT)
    ROW_RE: typing.Pattern = re.compile(r'^' + ROW_PAT + r'$')

    SLASH_HASH_PAT: str = r'/|#'
    SLASH_HASH_RE: typing.Pattern = re.compile(SLASH_HASH_PAT)

    # Instance attributes:

    # TODO: write a validator:
    input_file_paths: typing.List[Path] = attr.ib()

    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reject_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    namespace_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    updated_namespace_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    reader_options: KgtkReaderOptions = attr.ib(validator=attr.validators.instance_of(KgtkReaderOptions))

    allow_lax_uri: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_LAX_URI)

    allow_unknown_datatype_iris: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_UNKNOWN_DATATYPE_IRIS)

    allow_turtle_quotes: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_TURTLE_QUOTES)

    allow_lang_string_datatype: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_ALLOW_LANG_STRING_DATATYPE)
    lang_string_tag: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_LANG_STRING_TAG)

    local_namespace_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_LOCAL_NAMESPACE_PREFIX)
    local_namespace_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_LOCAL_NAMESPACE_USE_UUID)

    namespace_id_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_NAMESPACE_ID_PREFIX)
    namespace_id_counter: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NAMESPACE_ID_COUNTER)
    namespace_id_zfill: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NAMESPACE_ID_ZFILL)
    namespace_id_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_NAMESPACE_ID_USE_UUID)
    output_only_used_namespaces: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_OUTPUT_ONLY_USED_NAMESPACES)
    build_new_namespaces: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_BUILD_NEW_NAMESPACES)

    prefix_expansion_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_PREFIX_EXPANSION)

    structured_value_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STRUCTURED_VALUE)
    structured_uri_label: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STRUCTURED_URI)

    newnode_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_NEWNODE_PREFIX)
    newnode_counter: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NEWNODE_COUNTER)
    newnode_zfill: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_NEWNODE_ZFILL)
    newnode_use_uuid: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_NEWNODE_USE_UUID)

    build_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_BUILD_ID)
    idbuilder_options: typing.Optional[KgtkIdBuilderOptions] = attr.ib(default=None)
    idbuilder: typing.Optional[KgtkIdBuilder] = attr.ib(default=None)

    validate: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_VALIDATE)
    summary: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_SUMMARY)
    value_options: KgtkValueOptions = attr.ib(validator=attr.validators.instance_of(KgtkValueOptions), default=DEFAULT_KGTK_VALUE_OPTIONS)

    override_uuid: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    write_namespaces: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_WRITE_NAMESPACES)

    build_datatype_column: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_BUILD_DATATYPE_COLUMN)
    datatype_column_name: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_DATATYPE_COLUMN_NAME)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    local_namespace_uuid: str = attr.ib(default="")

    namespace_prefixes: typing.MutableMapping[str, str] = attr.ib(factory=dict)
    namespace_ids: typing.MutableMapping[str, str] = attr.ib(factory=dict)
    used_namespaces: typing.Set[str] = attr.ib(factory=set)

    # These get set as needed:
    converted_string_datatype_uri: str = attr.ib(default="")
    converted_lang_string_datatype_uri: str = attr.ib(default="")
    converted_number_datatype_uri: str = attr.ib(default="")

    output_line_count: int = attr.ib(default=0)
    unknown_datatype_iri_count: int = attr.ib(default=0)
    rejected_lang_string_count: int = attr.ib(default=0)

    def write_row(self, ew: KgtkWriter, node1: str, label: str, node2: str, datatype: str):
        output_row: typing.List[str]
        if self.build_datatype_column:
            output_row = [ node1, label, node2, datatype]
        else:
            output_row = [ node1, label, node2]

        if self.idbuilder is None:
            ew.write(output_row)
        else:
            ew.write(self.idbuilder.build(output_row, self.output_line_count))
        self.output_line_count += 1

    def convert_blank_node(self, item: str)->typing.Tuple[str, bool]:
        body: str = item[1:] # Strip the leading underscore, keep the colon.
        if self.local_namespace_use_uuid:
            return self.local_namespace_prefix + self.local_namespace_uuid + body, True
        else:
            return self.local_namespace_prefix + body, True

    def convert_uri(self, item: str, line_number: int)->typing.Tuple[str, bool]:
        body: str = item[1:-1] # Strip off the enclosing brackets.

        # First, check for an exact match for the body:
        namespace_id: str
        if body in self.namespace_prefixes:
            # Yes, this prefix exactly matches the body.
            namespace_id = self.namespace_prefixes[body]

            if self.output_only_used_namespaces:
                self.used_namespaces.add(namespace_id)

            # Return the namespace-prefixed URI:
            return namespace_id + ":", True

        # Move past "http://" or "https://":
        after_slashslash: int
        slashslash: int =  body.rfind("://")
        if slashslash < 1:
            if not self.allow_lax_uri:
                if self.verbose:
                    print("Line %d: invalid URI: '%s'" % (line_number, item), file=self.error_file, flush=True)
                return item, False
            after_slashslash = 0
        else:
            after_slashslash = slashslash + len("://")

        # Build a left-to-right list of slash- or hash-terminated sections:
        matches: typing.List[int] = [ ]
        m: typing.Match
        for m in self.SLASH_HASH_RE.finditer(body, after_slashslash):
            matches.append(m.end(0))

        # Search right-to-left for the longest match:
        matches.reverse()
        namespace_prefix: str
        suffix: str
        match_end: int
        for match_end in matches:
            namespace_prefix = body[:match_end]
            suffix = body[match_end:]
            if namespace_prefix in self.namespace_prefixes:
                # We have a winner.
                namespace_id = self.namespace_prefixes[namespace_prefix]

                if self.output_only_used_namespaces:
                    self.used_namespaces.add(namespace_id)

                # Return the namespace-prefixed URI:
                return namespace_id + ":" + suffix, True

        if not self.build_new_namespaces:
            # Do not attempt to build a new namespace.
            return body, True

        # Take the longest possible section, which is now first in the list:
        if len(matches) > 0:
            match_end = matches[0]
            namespace_prefix = body[:match_end]
            suffix = body[match_end:]
        else:
            # Take the entire body:
            namespace_prefix = body
            suffix = ""

        # Build a non-colliding namespace ID.
        while True:
            namespace_id = self.namespace_id_prefix
            if self.namespace_id_use_uuid:
                namespace_id += self.local_namespace_uuid + "-"
            namespace_id += str(self.namespace_id_counter).zfill(self.namespace_id_zfill)
            self.namespace_id_counter += 1
            if namespace_id not in self.namespace_ids:
                # Save the namespace ID for later reuse:
                self.namespace_ids[namespace_id] = namespace_prefix
                self.namespace_prefixes[namespace_prefix] = namespace_id

                if self.output_only_used_namespaces:
                    self.used_namespaces.add(namespace_id)

                # Return the namespace-prefixed URI:
                return namespace_id + ":" + suffix, True

    def convert_string(self, item: str, line_number: int)->typing.Tuple[str, bool]:
        # Convert this to a KGTK string.

        s: str = ast.literal_eval(item)
        return KgtkFormat.stringify(s), True
 
    def convert_lq_string(self, item: str, line_number: int)->typing.Tuple[str, bool]:
        # Convert this to a KGTK language qualified string.

        # Split the language qualifier (and optional suffix).
        # This code was copied from KgtkFormat.
        #
        # TODO: There should be only a single copy of this code.
        quoted_string: str
        language: str
        quoted_string, language = item.rsplit("@", 1)
        language_suffix: str = ""
        if "-" in language:
            language, language_suffix = language.split("-", 1)
            language_suffix = "-" + language_suffix

        # Parse the string, processing quoted characters:
        #
        # TODO: check for an error here!
        s: str = ast.literal_eval(quoted_string)

        # Assemble the final language-qualified string:
        return KgtkFormat.stringify(s, language, language_suffix), True

 
    def generate_new_node_symbol(self)->str:
        new_node_symbol: str = self.newnode_prefix
        if self.newnode_use_uuid:
            new_node_symbol += self.local_namespace_uuid + "-"
        new_node_symbol += str(self.newnode_counter).zfill(self.newnode_zfill)
        self.newnode_counter += 1
        return new_node_symbol
    
    def convert_boolean(self, item: str, value: str, line_number: int)->typing.Tuple[str, bool]:
        if value == 'true' or value == '1':
            return KgtkFormat.TRUE_SYMBOL, True
        elif value == 'false' or value == '0':
            return KgtkFormat.FALSE_SYMBOL, True
        else:
            if self.verbose:
                print("Line %d: invalid boolean item '%s'>" % (line_number, item), file=self.error_file, flush=True)
            return item, False
        

    # https://www.w3.org/2011/rdf-wg/wiki/XSD_Datatypes
    # https://www.w3.org/TR/xmlschema-2/

    NUMERIC_XSD_DATATYPES: typing.Set[str] = {
        '<http://www.w3.org/2001/XMLSchema#decimal>',
        '<http://www.w3.org/2001/XMLSchema#integer>',
        '<http://www.w3.org/2001/XMLSchema#int>',
        '<http://www.w3.org/2001/XMLSchema#short>',
        '<http://www.w3.org/2001/XMLSchema#byte>',
        '<http://www.w3.org/2001/XMLSchema#nonNegativeInteger>',
        '<http://www.w3.org/2001/XMLSchema#positiveInteger>',
        '<http://www.w3.org/2001/XMLSchema#unsignedLong>',
        '<http://www.w3.org/2001/XMLSchema#unsignedInt>',
        '<http://www.w3.org/2001/XMLSchema#unsignedShort>',
        '<http://www.w3.org/2001/XMLSchema#unsignedByte>',
        '<http://www.w3.org/2001/XMLSchema#nonPositiveInteger>',
        '<http://www.w3.org/2001/XMLSchema#negativeInteger>',
        '<http://www.w3.org/2001/XMLSchema#double>',
        '<http://www.w3.org/2001/XMLSchema#float>',
    }

    STRING_XSD_DATATYPES: typing.Set[str] = {
        '<http://www.w3.org/2001/XMLSchema#string>',
        '<http://www.w3.org/2001/XMLSchema#normalizedString>',
        '<http://www.w3.org/2001/XMLSchema#token>',
        '<http://www.w3.org/2001/XMLSchema#language>',
        '<http://www.w3.org/2001/XMLSchema#Name>',
        '<http://www.w3.org/2001/XMLSchema#NCName>',
        '<http://www.w3.org/2001/XMLSchema#ENTITY>',
        '<http://www.w3.org/2001/XMLSchema#ID>',
        '<http://www.w3.org/2001/XMLSchema#IDREF>',
        '<http://www.w3.org/2001/XMLSchema#NMTOKEN>',
    }

    LANG_STRING_DATATYPE_IRI: str = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#langString>"
    STRING_DATATYPE_IRI: str = '<http://www.w3.org/2001/XMLSchema#string>'
    NUMBER_DATATYPE_IRI: str = '<http://www.w3.org/2001/XMLSchema#number>'
                                                 
    def get_converted_string_datatype_uri(self)->str:
        if len(self.converted_string_datatype_uri) == 0 and self.build_datatype_column:
            # Convert the string datatype IRI and cache it:
            converted_string_datatype_uri_is_valid: bool
            self.converted_string_datatype_uri, converted_string_datatype_uri_is_valid = self.convert_uri(self.STRING_DATATYPE_IRI, 0)
            # TODO: fail if converted_string_datatype_uri_is_valid is False
        return self.converted_string_datatype_uri

    def get_converted_lang_string_datatype_uri(self)->str:
        if len(self.converted_lang_string_datatype_uri) == 0 and self.build_datatype_column:
            # Convert the langString datatype IRI and cache it:
            converted_lang_string_datatype_uri_is_valid: bool
            self.converted_lang_string_datatype_uri, converted_lang_string_datatype_uri_is_valid = self.convert_uri(self.LANG_STRING_DATATYPE_IRI, 0)
            # TODO: fail if converted_lang_string_datatype_uri_is_valid is False
        return self.converted_lang_string_datatype_uri

    def get_converted_number_datatype_uri(self)->str:
        if len(self.converted_number_datatype_uri) == 0 and self.build_datatype_column:
            # Convert the number datatype IRI and cache it:
            converted_number_datatype_uri_is_valid: bool
            self.converted_number_datatype_uri, converted_number_datatype_uri_is_valid = self.convert_uri(self.NUMBER_DATATYPE_IRI, 0)
            # TODO: fail if converted_number_datatype_uri_is_valid is False
        return self.converted_number_datatype_uri

    def convert_structured_literal(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool, str]:
        # This is the subset of strictured literals that fits the
        # pattern "STRING"^^<URI>.

        # Start by splitting on '^^'. We are certain it exists, and that the rightmost
        # instance is the one we want.
        uparrows: int = item.rfind("^^")
        if uparrows < 0:
            # This shouldn't happen!
            if self.verbose:
                print("Line %d: no uparrows in '%s'." % (line_number, item), file=self.error_file, flush=True)
            return item, False, ""

        string: str = item[:uparrows]
        uri: str = item[uparrows+2:]

        converted_uri: str = ""
        valid_uri: bool = True

        if uri in self.STRING_XSD_DATATYPES:
            # Convert this to a KGTK string.
            if self.build_datatype_column:
                converted_uri, valid_uri = self.convert_uri(uri, line_number)
                if not valid_uri:
                    return item, False, ""
            return self.convert_string(string, line_number) + (converted_uri,)

        elif uri in self.NUMERIC_XSD_DATATYPES:
            # Convert this to a KGTK number:
            if self.build_datatype_column:
                converted_uri, valid_uri = self.convert_uri(uri, line_number)
                if not valid_uri:
                    return item, False, ""
            return string[1:-1], True, converted_uri

        elif uri == '<http://www.w3.org/2001/XMLSchema#boolean>':
            # Convert this to a KGTK boolean:
            if self.build_datatype_column:
                converted_uri, valid_uri = self.convert_uri(uri, line_number)
                if not valid_uri:
                    return item, False, ""
            return self.convert_boolean(item, string[1:-1], line_number) + (converted_uri,)

        elif uri == '<http://www.w3.org/2001/XMLSchema#dateTime>':
            # Convert this to a KGTK date-and-time:
            #
            # Note: the W3C XML Schema standard allows the now obsolete
            # end-of-day time "24:00:00".
            if self.build_datatype_column:
                converted_uri, valid_uri = self.convert_uri(uri, line_number)
                if not valid_uri:
                    return item, False, ""
            return '^' + string[1:-1], True, converted_uri

        # TODO: the "date" schema
        # Problem:  it allows timezone offsets after dates without times!

        # Exposed langString datatypes are forbidden by the RDF 1.1 N-Triples
        # specification (and by the RDF 1.1 Turtle specification), but they
        # may occur in the wild anyway.  If we are so inclined, transform the
        # literal to an ordinary KGTK string.
        if uri == self.LANG_STRING_DATATYPE_IRI:
            if self.allow_lang_string_datatype:
                if len(self.lang_string_tag) == 0 or self.lang_string_tag == self.LANG_STRING_TAG_NONE:
                    # Convert this to a KGTK string.
                    return self.convert_string(string, line_number) + (self.get_converted_string_datatype_uri(),)
                else:
                    # Convert this to a KGTK language-qualified string.
                    return self.convert_lq_string(string + "@" + self.lang_string_tag, line_number) + (self.get_converted_lang_string_datatype_uri(),)
            else:
                self.rejected_lang_string_count += 1
                return item, False, ""

        if self.allow_unknown_datatype_iris:
            self.unknown_datatype_iri_count += 1

            converted_uri, valid_uri = self.convert_uri(uri, line_number)
            if not valid_uri:
                return item, False, ""

            if self.build_datatype_column:
                return string, True, converted_uri
            else:
                new_node_symbol: str = self.generate_new_node_symbol()
                self.write_row(ew, new_node_symbol, self.structured_value_label, string, "")
                self.write_row(ew, new_node_symbol, self.structured_uri_label, converted_uri, "")
                return new_node_symbol, True, ""
        
        # Give up on this unrecognized structured literal.
        return item, False, ""

    def convert_numeric(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool]:
        return item, True

    def convert(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool, str]:
        """
        Convert an ntriples item to KGTK format.

        TODO: update output_line_count for row written here.
        """
        if item.startswith("_:"):
            return self.convert_blank_node(item) + ("",)
        elif item.startswith("<") and item.endswith(">"):
            return self.convert_uri(item, line_number) + ("",)
        elif item.startswith('"') and item.endswith('"'):
            return self.convert_string(item, line_number) + (self.get_converted_string_datatype_uri(),)
        elif item.startswith('"') and item.endswith(">"):
            return self.convert_structured_literal(item, line_number, ew)
        elif item.startswith('"'):
            return self.convert_lq_string(item, line_number) + (self.get_converted_lang_string_datatype_uri(),)
        elif self.allow_turtle_quotes:
            if item.startswith("'") and item.endswith("'"):
                return self.convert_string(item, line_number) + (self.get_converted_string_datatype_uri(),)
            elif item.startswith("'") and item.endswith(">"):
                return self.convert_structured_literal(item, line_number, ew)
            elif item.startswith("'"):
                return self.convert_lq_string(item, line_number) + (self.get_converted_lang_string_datatype_uri(),)
        elif item[0] in "+-0123456789.":
            return self.convert_numeric(item, line_number, ew) + (self.get_converted_number_datatype_uri(),)

        if self.verbose:
            print("Line %d: unrecognized item '%s'" %(line_number, item), file=self.error_file, flush=True)

        return item, False, ""
    
    def convert_and_validate(self, item: str, line_number: int, ew: KgtkWriter)->typing.Tuple[str, bool, str]:
        result: str
        is_ok: bool
        datatype: str
        result, is_ok, datatype = self.convert(item, line_number, ew)

        # Just a little bit of paranoia here regarding tabs and ends-of-lines:
        #
        # TODO: perform these checks (and repairs!) in KgtkValue.
        if "\t" in result:
            result = result.replace("\t", "\\t")
        if "\n" in result:
            result = result.replace("\n", "\\n")
        if "\r" in result:
            result = result.replace("\r", "\\r")

        if is_ok and self.validate:
            kv: KgtkValue = KgtkValue(result, options=self.value_options)
            if not kv.validate():
                if self.verbose:
                    print("Input line %d: imported value '%s' (from '%s') is invalid." % (line_number, result, item),
                          file=self.error_file, flush=True)
                return result, False, ""
        return result, is_ok, datatype
            

    def get_default_namespaces(self)->int:
        namespace_id: str = "xml-schema-type"
        namespace_prefix: str = "http://www.w3.org/2001/XMLSchema#"
        self.namespace_ids[namespace_id] = namespace_prefix
        self.namespace_prefixes[namespace_prefix] = namespace_id
        return 1
    
    def get_initial_namespaces(self)->int:
        # Read the namespaces.  If no file, use a limited internal
        # default.
        if self.namespace_file_path is None:
            return self.get_default_namespaces()

        if self.verbose:
            print("Processing namespace file %s" % repr(str(self.namespace_file_path)), file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.namespace_file_path,
                                          mode=KgtkReaderMode.EDGE,
                                          options=self.reader_options,
                                          error_file=self.error_file,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )
        namespace_line_count: int = 0
        namespace_row: typing.List[str]
        for namespace_row in kr:
            namespace_line_count += 1
            if namespace_row[kr.label_column_idx] == self.prefix_expansion_label:
                namespace_id: str = namespace_row[kr.node1_column_idx]
                namespace_prefix: str = namespace_row[kr.node2_column_idx]
                if not (namespace_prefix.startswith('"') and namespace_prefix.endswith('"')):
                    if self.verbose:
                        print("The namespace prefix must be a KGKT string: '%s'" % namespace_prefix, file=self.error_file, flush=True)
                    continue
                
                # Strip the delimiting double quotes from the KGTk string.
                # Per RFC 3986, internal double quotes are not allowed in
                # a URL unless percent-encoded, so we needen't bother looking
                # for them.
                namespace_prefix = namespace_prefix[1:-1]

                if namespace_prefix in self.namespace_prefixes:
                    if self.verbose:
                        print("Duplicate initial namespace prefix '%s'" % namespace_prefix, file=self.error_file, flush=True)
                else:
                    self.namespace_prefixes[namespace_prefix] = namespace_id
                if namespace_id in self.namespace_ids:
                    if self.verbose:
                        print("Duplicate initial namespace id '%s'" % namespace_id, file=self.error_file, flush=True)
                else:
                    self.namespace_ids[namespace_id] = namespace_prefix
            else:
                if self.verbose:
                    print("Ignoring initial namespace label '%s'" % namespace_row[kr.label_column_idx], file=self.error_file, flush=True)

        return namespace_line_count
        
    def write_namespaces_to_output(self, ew: KgtkWriter, count_records: bool = True):
        if self.write_namespaces:
            # Append the namespaces to the output file.
            namespace_id: str
            for namespace_id in sorted(self.namespace_ids.keys()):
                if self.output_only_used_namespaces:
                    if namespace_id in self.used_namespaces:
                        self.write_row(ew, namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"', "")
                else:
                    self.write_row(ew, namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"', "")
 
    def write_updated_namespace_file(self):
        # Is there an updated namespaces file?
        if self.updated_namespace_file_path is None:
            return

        if self.verbose:
            print("Opening updated namespaces file %s" % str(self.updated_namespace_file_path), file=self.error_file, flush=True)
        # Open the updated namespaces file.
        un: KgtkWriter = KgtkWriter.open(self.COLUMN_NAMES,
                                         self.updated_namespace_file_path,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)
        namespace_id: str
        for namespace_id in sorted(self.namespace_ids.keys()):
            un.write([namespace_id, self.prefix_expansion_label, '"' + self.namespace_ids[namespace_id] + '"'])
        un.close()


    def save_namespaces(self, ew: KgtkWriter):
        self.write_namespaces_to_output(ew)
        self.write_updated_namespace_file()

    def parse(self, line: str, line_number: int)->typing.Tuple[typing.List[str], bool, bool]:
        m: typing.Optional[typing.Match] = self.ROW_RE.match(line)
        if m is None:
            if self.verbose:
                print("Line %d: not parsed.\n%s" % (line_number, line), file=self.error_file, flush=True)
            return [ ], False, False
        return [m.group("node1"), m.group("label"), m.group("node2")], True, len(m.group("comment")) > 0


    def process(self):
        output_column_names: typing.List[str] = self.COLUMN_NAMES.copy()
        if self.build_datatype_column:
            output_column_names.append(self.datatype_column_name)
        if self.build_id and self.idbuilder_options is not None:
            self.idbuilder = KgtkIdBuilder.from_column_names(output_column_names, self.idbuilder_options)
            output_column_names = self.idbuilder.column_names

        if self.verbose:
            print("Opening output file %s" % str(self.output_file_path), file=self.error_file, flush=True)
        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        rw: typing.Optional[typing.TextIO] = None
        if self.reject_file_path is not None:
            if self.verbose:
                print("Opening reject file %s" % repr(str(self.reject_file_path)), file=self.error_file, flush=True)
            # Open the reject file. Since the input data is not in KGTK format,
            # we use an ordinary file here.
            if str(self.reject_file_path) == "-":
                rw = sys.stdout
            else:
                rw = open(self.reject_file_path, "wt")


        total_input_line_count: int = 0
        reject_line_count: int = 0
        comment_count: int = 0
        
        namespace_line_count: int = self.get_initial_namespaces()

        input_file_path: str
        for input_file_path in self.input_file_paths:
            input_line_count: int = 0
            
            if self.local_namespace_use_uuid or self.namespace_id_use_uuid or self.newnode_use_uuid:
                if self.override_uuid is not None:
                    self.local_namespace_uuid = self.override_uuid # for debugging
                else:
                    # Generate a new local namespace UUID.
                    self.local_namespace_uuid = shortuuid.uuid()

            # Open the input file.
            if self.verbose:
                print("Opening the input file: %s" % input_file_path, file=self.error_file, flush=True)
            infile: typing.TextIO
            if str(input_file_path) == "-":
                infile = sys.stdin
            else:
                infile = open(input_file_path, 'rt') 

            line: str
            for line in infile:
                input_line_count += 1
                total_input_line_count += 1

                row: typing.List[str]
                valid: bool
                has_comment: bool
                row, valid, has_comment = self.parse(line, input_line_count)
                if not valid:
                    if rw is not None:
                        rw.write(line)
                    reject_line_count += 1
                    continue

                node1: str
                ok_1: bool
                node1, ok_1, _ = self.convert_and_validate(row[0], input_line_count, ew)

                label: str
                ok_2: bool
                label, ok_2, _ = self.convert_and_validate(row[1], input_line_count, ew)

                node2: str
                ok_3: bool
                datatype: str
                node2, ok_3, datatype = self.convert_and_validate(row[2], input_line_count, ew)

                if ok_1 and ok_2 and ok_3:
                    self.write_row(ew, node1, label, node2, datatype)
                else:
                    if rw is not None:
                        rw.write(line)
                    reject_line_count += 1

                if has_comment:
                    comment_count += 1

            if input_file_path != "-":
                infile.close()


        self.save_namespaces(ew)

        if self.verbose or self.summary:
            print("Processed %d known namespaces." % (namespace_line_count), file=self.error_file, flush=True)
            print("Processed %d records." % (total_input_line_count), file=self.error_file, flush=True)
            print("Rejected %d records." % (reject_line_count), file=self.error_file, flush=True)
            print("Wrote %d records." % (self.output_line_count), file=self.error_file, flush=True)
            print("Ignored %d comments." % (comment_count), file=self.error_file, flush=True)
            print("Rejected %d records with langString IRIs." % (self.rejected_lang_string_count), file=self.error_file, flush=True)
            print("Imported %d records with unknown datatype IRIs." % (self.unknown_datatype_iri_count), file=self.error_file, flush=True)

        if ew is not None:
            ew.close()
            
        if rw is not None and self.reject_file_path is not None and self.reject_file_path != "-":
            rw.close()
            
    @classmethod
    def add_arguments(cls, parser: ArgumentParser):

        parser.add_argument(      "--namespace-id-prefix", dest="namespace_id_prefix",
                                  help="The prefix used to generate new namespaces. (default=%(default)s).",
                                  default=cls.DEFAULT_NAMESPACE_ID_PREFIX)
    
        parser.add_argument(      "--namespace-id-use-uuid", dest="namespace_id_use_uuid",
                                  help="Use the local namespace UUID when generating namespaces. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_NAMESPACE_ID_USE_UUID)

        parser.add_argument(      "--namespace-id-counter", dest="namespace_id_counter",
                                  help="The counter used to generate new namespaces. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NAMESPACE_ID_COUNTER)
    
        parser.add_argument(      "--namespace-id-zfill", dest="namespace_id_zfill",
                                  help="The width of the counter used to generate new namespaces. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NAMESPACE_ID_ZFILL)
    
        parser.add_argument(      "--output-only-used-namespaces", dest="output_only_used_namespaces",
                                  help="Write only used namespaces to the output file. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_OUTPUT_ONLY_USED_NAMESPACES)

        parser.add_argument(      "--build-new-namespaces", dest="build_new_namespaces",
                                  help="When True, create new namespaces.  When False, use only existing namespaces. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_BUILD_NEW_NAMESPACES)

        parser.add_argument(      "--allow-lax-uri", dest="allow_lax_uri",
                                  help="Allow URIs that don't begin with a http:// or https://. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_LAX_URI)

        parser.add_argument(      "--allow-unknown-datatype-iris", dest="allow_unknown_datatype_iris",
                                  help="Allow unknown datatype IRIs, creating  a qualified record. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_UNKNOWN_DATATYPE_IRIS)

        parser.add_argument(      "--allow-turtle-quotes", dest="allow_turtle_quotes",
                                  help="Allow literals to use single quotes (to support Turtle format). (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_TURTLE_QUOTES)

        parser.add_argument(      "--allow-lang-string-datatype", dest="allow_lang_string_datatype",
                                  help="Allow literals to include exposed langString datatype IRIs (which is forbidden by the spec, but occurs anyway). (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_ALLOW_LANG_STRING_DATATYPE)

        parser.add_argument(      "--lang-string-tag", dest="lang_string_tag",
                                  help="The tag to use with exposed langString instances. `` or `-` mean to use a string, otherwise use a lanuage-qualified string. (default=%(default)s).",
                                  default=cls.DEFAULT_LANG_STRING_TAG)
    
        parser.add_argument(      "--local-namespace-prefix", dest="local_namespace_prefix",
                                  help="The namespace prefix for blank nodes. (default=%(default)s).",
                                  default=cls.DEFAULT_LOCAL_NAMESPACE_PREFIX)

        parser.add_argument(      "--local-namespace-use-uuid", dest="local_namespace_use_uuid",
                                  help="Generate a UUID for the local namespace. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_LOCAL_NAMESPACE_USE_UUID)

        parser.add_argument(      "--prefix-expansion-label", dest="prefix_expansion_label",
                                  help="The label for prefix expansions in the namespace file. (default=%(default)s).",
                                  default=cls.DEFAULT_PREFIX_EXPANSION)
    
        parser.add_argument(      "--structured-value-label", dest="structured_value_label",
                                  help="The label for value records for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_STRUCTURED_VALUE)
    
        parser.add_argument(      "--structured-uri-label", dest="structured_uri_label",
                                  help="The label for URI records for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_STRUCTURED_URI)
    
        parser.add_argument(      "--newnode-prefix", dest="newnode_prefix",
                                  help="The prefix used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  default=cls.DEFAULT_NEWNODE_PREFIX)
    
        parser.add_argument(      "--newnode-use-uuid", dest="newnode_use_uuid",
                                  help="Use the local namespace UUID when generating new nodes for ntriple structured literals. " +
                                  "When there are multiple input files, each input file gets its own UUID. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_NEWNODE_USE_UUID)

        parser.add_argument(      "--newnode-counter", dest="newnode_counter",
                                  help="The counter used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NEWNODE_COUNTER)
    
        parser.add_argument(      "--newnode-zfill", dest="newnode_zfill",
                                  help="The width of the counter used to generate new nodes for ntriple structured literals. (default=%(default)s).",
                                  type=int, default=cls.DEFAULT_NEWNODE_ZFILL)
    
        parser.add_argument(      "--build-id", dest="build_id",
                                  help="Build id values in an id column. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_BUILD_ID)

        parser.add_argument(      "--build-datatype-column", dest="build_datatype_column",
                                  help="When True, and --datatype-column-name DATATYPE_COLUMN_NAME is not empty, build a column with RDF datatypes. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_BUILD_DATATYPE_COLUMN)

        parser.add_argument(      "--datatype-column-name", dest="datatype_column_name",
                                  help="The name of the column with RDF datatypes. (default=%(default)s).",
                                  default=cls.DEFAULT_DATATYPE_COLUMN_NAME)
    
        parser.add_argument(      "--validate", dest="validate",
                                  help="When true, validate that the result fields are good KGTK file format. (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_VALIDATE)

        parser.add_argument(      "--summary", dest="summary",
                                  help="When true, print summary statistics when done processing (also implied by --verbose). (default=%(default)s).",
                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_SUMMARY)

        parser.add_argument(      "--override-uuid", dest="override_uuid",
                                  help="When specified, override UUID generation for debugging. (default=%(default)s).",
                                  default=None)

        parser.add_argument(      "--write-namespaces", dest="write_namespaces",
                                  help="When true, append namespaces to the output file. (default=%(default)s).",
                                  default=cls.DEFAULT_WRITE_NAMESPACES)

def main():
    """
    Test the KGTK ntriples importer.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument("-i", "--input-files", dest="input_file_paths", nargs='*',
                        help="The file(s) with the input ntriples data. (default=%(default)s)", type=Path, default="-")

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--reject-file", dest="reject_file_path", help="The KGTK file into which to write rejected records. (default=%(default)s).",
                              type=Path, default=None)
    
    parser.add_argument(      "--namespace-file", dest="namespace_file_path", help="The KGTK file with known namespaces. (default=%(default)s).",
                              type=Path, default=None)
    
    parser.add_argument(      "--updated-namespace-file", dest="updated_namespace_file_path",
                              help="An updated KGTK file with known namespaces. (default=%(default)s).",
                              type=Path, default=None)
    

    KgtkNtriples.add_arguments(parser)
    KgtkIdBuilderOptions.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_args(args)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("--input-files %s" % " ".join([repr(str(path)) for path in input_file_paths]), file=error_file, flush=True)
        print("--output-file=%s" % repr(str(args.output_file_path)), file=error_file, flush=True)
        # TODO: show ifempty-specific options.
        if args.reject_file_path is not None:
            print("--reject-file=%s" % repr(str(args.reject_file_path)), file=error_file, flush=True)
        if args.namespace_file_path is not None:
            print("--namespace-file=%s" % repr(str(args.namespace_file_path)), file=error_file, flush=True)
        if args.updated_namespace_file_path is not None:
            print("--updated-namespace-file=%s" % repr(str(args.updated_namespace_file_path)), file=error_file, flush=True)
        print("--namespace-id-prefix %s" % repr(args.namespace_id_prefix), file=error_file, flush=True)
        print("--namespace-id-use-uuid %s" % repr(args.namespace_id_use_uuid), file=error_file, flush=True)
        print("--namespace-id-counter %s" % repr(args.namespace_id_counter), file=error_file, flush=True)
        print("--namespace-id-zfill %s" % repr(args.namespace_id_zfill), file=error_file, flush=True)
        print("--output-only-used-namespaces %s" % repr(args.output_only_used_namespaces), file=error_file, flush=True)
        print("--build-new-namespaces %s" % repr(args.build_new_namespaces), file=error_file, flush=True)
        print("--allow-lax-uri %s" % repr(args.allow_lax_uri), file=error_file, flush=True)
        print("--allow-unknown-datatype-iris %s" % repr(args.allow_unknown_datatype_iris), file=error_file, flush=True)
        print("--allow-turtle-quotes %s" % repr(args.allow_turtle_quotes), file=error_file, flush=True)
        print("--allow-lang-string-datatype %s" % repr(args.allow_lang_string_datatype), file=error_file, flush=True)
        print("--lang-string-tag %s" % repr(args.lang_string_tag), file=error_file, flush=True)
        print("--local-namespace-prefix %s" % repr(args.local_namespace_prefix), file=error_file, flush=True)
        print("--local-namespace-use-uuid %s" % repr(args.local_namespace_use_uuid), file=error_file, flush=True)
        print("--prefix-expansion-label %s" % repr(args.prefix_expansion_label), file=error_file, flush=True)
        print("--structured-value-label %s" % repr(args.structured_value_label), file=error_file, flush=True)
        print("--structured-uri-label %s" % repr(args.structured_uri_label), file=error_file, flush=True)
        print("--newnode-prefix %s" % repr(args.newnode_prefix), file=error_file, flush=True)
        print("--newnode-use-uuid %s" % repr(args.newnode_use_uuid), file=error_file, flush=True)
        print("--newnode-counter %s" % repr(args.newnode_counter), file=error_file, flush=True)
        print("--newnode-zfill %s" % repr(args.newnode_zfill), file=error_file, flush=True)
        print("--build-id=%s" % repr(args.build_id), file=error_file, flush=True)
        print("--build-datatype-column %s" % repr(args.build_datatype_column), file=error_file, flush=True)
        print("--datatype-column-name %s" % repr(args.datatype_column_name), file=error_file, flush=True)
        print("--validate=%s" % repr(args.validate), file=error_file, flush=True)
        print("--summary=%s" % repr(args.summary), file=error_file, flush=True)
        if args.override_uuid is not None:
            print("--override_uuid=%s" % repr(args.override_uuid), file=error_file, flush=True)            
        print("--write-namespaces=%s" % repr(args.write_namespaces), file=error_file, flush=True)

        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kn: KgtkNtriples = KgtkNtriples(
        input_file_paths=args.input_file_paths,
        output_file_path=args.output_file_path,
        reject_file_path=args.reject_file_path,
        namespace_file_path=args.namespace_file_path,
        updated_namespace_file_path=args.updated_namespace_file_path,
        namespace_id_prefix=args.namespace_id_prefix,
        namespace_id_use_uuid=args.namespace_id_use_uuid,
        namespace_id_counter=args.namespace_id_counter,
        namespace_id_zfill=args.namespace_id_zfill,
        output_only_used_namespaces=args.output_only_used_namespaces,
        build_new_namespaces=args.build_new_namespaces,
        newnode_prefix=args.newnode_prefix,
        newnode_use_uuid=args.newnode_use_uuid,
        newnode_counter=args.newnode_counter,
        newnode_zfill=args.newnode_zfill,
        allow_lax_uri=args.allow_lax_uri,
        allow_unknown_datatype_iris=args.allow_unknown_datatype_iris,
        allow_turtle_quotes=args.allow_turtle_quotes,
        allow_lang_string_datatype=args.allow_lang_string_datatype,
        lang_string_tag=args.lang_string_tag,
        local_namespace_prefix=args.local_namespace_prefix,
        local_namespace_use_uuid=args.local_namespace_use_uuid,
        prefix_expansion_label=args.prefix_expansion_label,
        structured_value_label=args.structured_value_label,
        structured_uri_label=args.structured_uri_label,
        build_id=args.build_id,
        idbuilder_options=idbuilder_options,
        build_datatype_column=args.build_datatype_column,
        datatype_column_name=args.datatype_column_name,
        validate=args.validate,
        summary=args.summary,
        override_uuid=args.override_uuid,
        write_namespaces=args.write_namespaces,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    kn.process()
    
if __name__ == "__main__":
    main()
